{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "%pylab inline\n",
    "\n",
    "#### Load constants ####\n",
    "import scipy.constants as sc\n",
    "from matplotlib.colors import LogNorm\n",
    "import sys\n",
    "from time import time\n",
    "\n",
    "\n",
    "# OVERWRITE DEFAULT PLOTTING PARAMETERS\n",
    "params = {\n",
    "    'font.family' : 'sans-serif',\n",
    "    'font.sans-serif' : ['DejaVu Sans', 'Arial', 'Verdana', 'Liberation Sans'],\n",
    "    'mathtext.default' : 'regular',\n",
    "    'mathtext.rm' : 'sans',\n",
    "    'font.size' : 18,\n",
    "    'legend.fontsize' : 14,\n",
    "    'axes.labelsize' : 18,\n",
    "    'axes.titlesize' : 16,\n",
    "    'figure.titlesize' : 16,\n",
    "    'lines.linewidth' : 3.0,\n",
    "    'legend.frameon' : False,\n",
    "    'legend.numpoints': 1,\n",
    "}\n",
    "\n",
    "matplotlib.rcParams.update(params)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Implementation of laser pulses\n",
    "\n",
    "Different laser pulses are implemented as functors.\n",
    "That is, instances of these classes can be evaluated like functions after instantation."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Complex valued electric field of laser pulses in frequency-space domain\n",
    "\n",
    "### (1) Gaussian laser pulse"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class GaussPulseFourierSpace(object):\n",
    "    \"\"\" \n",
    "        Elecric field of a Gaussian laser pulse.\n",
    "        That is, a field with Gaussian frequency spectrum and Gaussian transverse envelope.\n",
    "        Individual frequencies are modeled as plane waves.\n",
    "        Compare eq. (66) in supplemental material of\n",
    "        K. Steiniger et al, \"Distortions in focusing laser pulses due to spatio-temporal couplings–An analytic description\".\n",
    "        \n",
    "        Implementation is for a pulse at arbitrary distance z from its focus.\n",
    "        \n",
    "        Before propagation, the pulse's electric field in the focus is assumed to be:\n",
    "            E(ν,x) = ϵ(ν-ν0) * exp{ -x^2/w0^2 } * exp{ -( 2π * SD * (ν-ν0) / w_0 )^2 }\n",
    "                     * exp{( α_2 + I*α_3 )*(x/w_0)}\n",
    "                     * exp{- I * .5 * GDD_0 * }\n",
    "        with\n",
    "            ϵ(ν) = .5 * sqrt{π} * τ_Laser exp{ -π^2 ν^2 * τ_Laser^2}\n",
    "            α_2 = 4π * SD * (ν-ν0) / w_0\n",
    "            α_3 = (w_0 / c) * 4π^2[ ν_0 * θ' * (ν-ν0) + .5 * ( 2*θ' + 2πν_0 * θ'' ) * (ν - ν0)^2 ]\n",
    "        \n",
    "        Compare with eq. (17) of the above article.\n",
    "            \n",
    "        Be aware, derivatives of the frequency dependent propagation angle θ (θ' = AD),\n",
    "        and distribution center x_0 (x_0' = SD_0), are with respect to\n",
    "        the angular frequency Ω, not ν.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self\n",
    "                 , wavelength # central wavelength of the laser, unit: m\n",
    "                 , fourier_limited_pulse_duration # where the field(!) drops to 1/e,\n",
    "                                                  # e.g. tau_FWHM,I / sqrt(2. * log(2.)),\n",
    "                                                  # unit: s\n",
    "                 , fourier_limited_width_in_focus # where the field(!) drops to 1/e,\n",
    "                                                  # e.g. w_FWHM,I / sqrt(2. * log(2.)),\n",
    "                                                  # unit: m\n",
    "                 , angular_dispersion_in_focus # dθ/dΩ, e.g. tan(α_tilt)/Ω0, unit: rad * s\n",
    "                 , spatial_dispersion_in_focus # dx/dΩ, unit: m*s\n",
    "                 , group_delay_dispersion_in_focus # (d^2φ/dΩ^2) unit: s^2\n",
    "                 , third_order_dispersion_in_focus = 0. # (d^3φ/dΩ^3) unit: s^3\n",
    "                ):\n",
    "        self.λ_Laser = wavelength\n",
    "        self.ν0 = sc.c / self.λ_Laser\n",
    "        self.Ω0 = 2. * sc.pi * self.ν0\n",
    "        self.τ0 = fourier_limited_pulse_duration\n",
    "        self.w0 = fourier_limited_width_in_focus\n",
    "        self.zR = sc.pi * self.w0**2 / self.λ_Laser #Rayleigh length\n",
    "        self.AD_foc = angular_dispersion_in_focus\n",
    "        self.SD_foc = spatial_dispersion_in_focus\n",
    "        self.GDD_foc = group_delay_dispersion_in_focus\n",
    "        self.TOD_foc = third_order_dispersion_in_focus\n",
    "    \n",
    "    \n",
    "    def __call__(self, ν, x, z, TD = 0.):\n",
    "        \"\"\" Return real and imaginary part of complex electric field in frequency-space domain\n",
    "            at given frequency ν and position (x,z), where z is\n",
    "            the laser propagation direction.\n",
    "            \n",
    "            Laser field has the form E = E_amp * e^[-I φ]\n",
    "\n",
    "            \n",
    "            ν - is an array of frequencies for which the spectral field is calculated\n",
    "        \"\"\"\n",
    "        return array([ self.amplitude(ν_i, x, z)\n",
    "                       * array([ cos(self.phase_angle(ν_i, x, z, TD))\n",
    "                                , - sin(self.phase_angle(ν_i, x, z, TD)) ])\n",
    "                      for ν_i in ν])\n",
    "    \n",
    "    \n",
    "    def amplitude(self, ν, x ,z):\n",
    "        \"\"\" Real spectral amplitude E_amp of the pulse's electric field.\n",
    "        \"\"\"\n",
    "        return ( self.spectrum(ν) * sqrt(self.w0 / self.w(z)) \n",
    "                 * exp( - ( ( x - self.x_c(ν,z) ) / self.w(z) )**2 )\n",
    "               )\n",
    "    \n",
    "    \n",
    "    def phase_angle(self, ν, x ,z, TD):\n",
    "        \"\"\" Spectral phase φ\n",
    "                    \n",
    "            Analytic expectation after propagation of plane waves along z with angular dispersion \n",
    "            including initial time delay, spatial dispersion and group-delay dispersion.\n",
    "        \"\"\"\n",
    "        return (+ 2. * sc.pi * ν * z / sc.c\n",
    "                + 2. * sc.pi * (ν - self.ν0) * TD\n",
    "                + ( x - self.x_c(ν,z) )**2 * .5 * 2.*sc.pi* ν * self.R_inv(z) / sc.c\n",
    "                - self.α_3(ν,z) * x / self.w0\n",
    "                - 0.25 * self.α_3(ν,z)**2 * z / self.zR\n",
    "                - .5 * arctan( z / self.zR )\n",
    "                + .5 * self.GDD_foc * ( 2. * sc.pi * (ν - self.ν0) )**2\n",
    "                + self.TOD_foc * ( 2. * sc.pi * (ν - self.ν0) )**3 / 6.\n",
    "               )\n",
    "        \n",
    "    \n",
    "    def spectrum(self,ν):\n",
    "        \"\"\" Gaussian spectrum yielding a pulse with an FWHM duration of the intensity of\n",
    "                tau_FWHM,I = sqrt(2.*ln(2.)) * tau0\n",
    "        \"\"\"\n",
    "        norm = sqrt(sc.pi) * self.τ0 \n",
    "        return norm * exp(- self.τ0**2 * sc.pi**2 * (ν - self.ν0)**2 )\n",
    "    \n",
    "    \n",
    "    def x_c(self, ν, z):\n",
    "        \"\"\" Center of a frequency's spatial distribution\n",
    "        \"\"\"\n",
    "        return ( 2. * sc.pi * self.SD_foc * (ν - self.ν0)\n",
    "                 - (sc.c / ( 2. * sc.pi * self.ν0 * self.w0 ) ) * self.α_3(ν,z) * z\n",
    "               )\n",
    "    \n",
    "    \n",
    "    def w(self,z):\n",
    "        \"\"\" Laser width without dispersion\n",
    "        \"\"\"\n",
    "        return self.w0 * sqrt( 1. + (z / self.zR)**2 )\n",
    "    \n",
    "    \n",
    "    def R_inv(self,z):\n",
    "        \"\"\" Inverse of typical laser curvature R(z) of Gaussian pulses, i.e. 1 / R(z)\n",
    "        \"\"\"\n",
    "        return z / ( z**2 + self.zR**2 )\n",
    "    \n",
    "    \n",
    "    def α_3(self,ν,z):\n",
    "        \"\"\" Initial frequency dependent complex phase expanded up to third\n",
    "            order in (Ω-Ω0).\n",
    "            Yet, this takes only first order angular dispersion dθ/dΩ=θ' into account\n",
    "            and neglects all higher order angular dispersion terms, e.g.\n",
    "            d^2θ/dΩ^2 = θ'', θ''', ...\n",
    "        \"\"\"\n",
    "        return ( \n",
    "            self.ν0 * self.AD_foc * (ν - self.ν0)\n",
    "            + self.AD_foc * (ν - self.ν0)**2\n",
    "            - 4. * sc.pi**2 * self.ν0 * self.AD_foc**3 * (ν - self.ν0)**3 / 6.\n",
    "        ) * 4. * sc.pi**2 * self.w0 / sc.c\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (2) Implementation of library-class which holds a lot of quantities required in the evaluation of the analytically propagated and Fourier-transformed laser pulses"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class FourierSpaceFunctionLibrary(object):\n",
    "    \"\"\" Implementation of functions appearing in the Fourier space complex electric field definition\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self\n",
    "                 , wavelength # central wavelength of the laser, unit: m\n",
    "                 , fourier_limited_pulse_duration # where the field(!) drops to 1/e,\n",
    "                                                  # e.g. tau_FWHM,I / sqrt(2. * log(2.)),\n",
    "                                                  # unit: s\n",
    "                 , fourier_limited_width_in_focus # where the field(!) drops to 1/e,\n",
    "                                                  # e.g. w_FWHM,I / sqrt(2. * log(2.)),\n",
    "                                                  # unit: m\n",
    "                 , angular_dispersion_in_focus # dθ/dΩ, e.g. tan(α_tilt)/Ω0, unit: rad * s\n",
    "                 , spatial_dispersion_in_focus # dx/dΩ, unit: m*s\n",
    "                 , group_delay_dispersion_in_focus # (d^2φ/dΩ^2) unit: s^2\n",
    "                ):\n",
    "        self.λ_Laser = wavelength\n",
    "        self.ν0 = sc.c / self.λ_Laser\n",
    "        self.Ω0 = 2. * sc.pi * self.ν0\n",
    "        self.τ0 = fourier_limited_pulse_duration\n",
    "        self.w0 = fourier_limited_width_in_focus\n",
    "        self.zR = sc.pi * self.w0**2 / self.λ_Laser #Rayleigh length\n",
    "        self.AD_foc = angular_dispersion_in_focus\n",
    "        self.SD_foc = spatial_dispersion_in_focus\n",
    "        self.GDD_foc = group_delay_dispersion_in_focus\n",
    "    \n",
    "    def w(self, z):\n",
    "        \"\"\" Laser width without dispersion\n",
    "        \"\"\"\n",
    "        return self.w0 * sqrt( 1. + (z / self.zR)**2 )\n",
    "    \n",
    "    def R_inv(self, z):\n",
    "        \"\"\" Inverse of laser curvature R(z), i.e. 1 / R(z)\n",
    "        \"\"\"\n",
    "        return z / ( z**2 + self.zR**2 )\n",
    "    \n",
    "    def beta_2(self, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        return ( z - self.Ω0 * self.AD_foc * x ) / sc.c\n",
    "    \n",
    "    def beta_3(self, z):\n",
    "        \"\"\" Auxialliary quantity.\n",
    "            See analytical derivation of PFT\n",
    "        \"\"\"\n",
    "        return ( self.AD_foc * z - self.SD_foc ) / self.w(z)\n",
    "    \n",
    "    def beta_4(self, z):\n",
    "        \"\"\" Assuming, there are no second order dispersions present\n",
    "        \"\"\"\n",
    "        return self.AD_foc * z / ( self.w(z) * self.Ω0 )\n",
    "    \n",
    "    def beta_5(self, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        return self.w(z)**2 * self.R_inv(z) / (2. * sc.c)\n",
    "    \n",
    "    def beta_6(self, x):\n",
    "        \"\"\" Be aware, that in contrast to the publication,\n",
    "            beta_6 from the publication is split into\n",
    "            beta_6 -> beta_6 + beta_7, here.\n",
    "        \"\"\"\n",
    "        return self.AD_foc * x / sc.c\n",
    "    \n",
    "    def beta_7(self, z):\n",
    "        \"\"\" This term is part of beta_6 in the publication,\n",
    "            since they always appear in a pair anyways.\n",
    "        \"\"\"\n",
    "        return self.Ω0 * self.AD_foc**2 * z / (2. * sc.c)\n",
    "    \n",
    "    def beta_8(self, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        assert(False)\n",
    "    \n",
    "    def beta_9(self, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        assert(False)\n",
    "    \n",
    "    def gamma_1(self, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        return ( 1.\n",
    "            + 8. * x * self.beta_4(z) / ( self.w(z) * self.τ0**2 )\n",
    "            + 4. * self.beta_3(z)**2 / self.τ0**2\n",
    "        ) \n",
    "    \n",
    "    def gamma_2(self, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        w = self.w(z)\n",
    "        β_3 = self.beta_3(z)\n",
    "        β_4 = self.beta_4(z)\n",
    "        β_5 = self.beta_5(z)\n",
    "        β_6 = self.beta_6(x)\n",
    "        β_7 = self.beta_7(z)\n",
    "        return ( .5 * self.GDD_foc - β_6 - β_7 + 2. * x * β_3 * β_5 / w\n",
    "               + ( 2. * x * β_4 / w + β_3**2 ) * self.Ω0 * β_5\n",
    "        ) * 4. / self.τ0**2\n",
    "    \n",
    "    def gamma_3(self, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        return -2. * x * self.beta_3(z) / ( self.w(z) * self.τ0 )\n",
    "    \n",
    "    def gamma_4(self, t, x, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        w = self.w(z)\n",
    "        β_5 = self.beta_5(z)\n",
    "        return ( t\n",
    "            - self.beta_2(x, z)\n",
    "            - β_5 * (x / w)**2 \n",
    "            - 2. * x * self.Ω0 * self.beta_3(z) * β_5 / w\n",
    "        ) / self.τ0\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (3) Gaussian laser pulse keeping only terms of order $(\\Omega-\\Omega_0)^2$ and lower"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class ApproximateGaussPulseFourierSpace(object):\n",
    "    \"\"\" Compare eq. (97) in supplemental material of\n",
    "        K. Steiniger et al, \"Distortions in focusing laser pulses due to spatio-temporal couplings–An analytic description\".\n",
    "        \n",
    "        Basically same as `GaussPulseFourierSpace`,\n",
    "        but all terms of order (Ω-Ω0)**3 and higher are neglected.\n",
    "        This is the pulse which is analytically Fourier transformed to\n",
    "        obtain relations for PFT and pulse duration.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self\n",
    "                 , wavelength\n",
    "                 , fourier_limited_pulse_duration = 30.e-15 / 1.17741 # tau_FWHM,I / sqrt(2. * log(2.))\n",
    "                 , fourier_limited_width_in_focus = 3.e-6 / 1.17741 # w_FWHM,I / sqrt(2. * log(2.))\n",
    "                 , angular_dispersion_in_focus = 0. # tan(α_tilt)/(2.*sc.pi*ν0)\n",
    "                 , spatial_dispersion_in_focus = 0.\n",
    "                 , group_delay_dispersion_in_focus = 0.\n",
    "                ):\n",
    "        self.funcLib = FourierSpaceFunctionLibrary(\n",
    "            wavelength\n",
    "            , fourier_limited_pulse_duration\n",
    "            , fourier_limited_width_in_focus\n",
    "            , angular_dispersion_in_focus\n",
    "            , spatial_dispersion_in_focus\n",
    "            , group_delay_dispersion_in_focus\n",
    "        )\n",
    "        self.λ_Laser = self.funcLib.λ_Laser\n",
    "        self.ν0 = self.funcLib.ν0\n",
    "        self.Ω0 = self.funcLib.Ω0\n",
    "        self.τ0 = self.funcLib.τ0\n",
    "        self.w0 = self.funcLib.w0\n",
    "        self.zR = self.funcLib.zR #Rayleigh length\n",
    "        self.AD_foc = self.funcLib.AD_foc\n",
    "        self.SD_foc = self.funcLib.SD_foc\n",
    "        self.GDD_foc = self.funcLib.GDD_foc\n",
    "    \n",
    "    \n",
    "    def __call__(self, ν, x, z, TD = 0.):\n",
    "        \"\"\" Return real and imaginary part of the pulse's complex electric field\n",
    "                E = E_amp * e^[-I φ]\n",
    "            at given frequency ν and position (x,z), where z is\n",
    "            the pulse's propagation direction.\n",
    "            \n",
    "            ν - is an array of frequencies for which the spectral field is calculated\n",
    "            TD - is a time delay in order to move the pulse's center position to a location\n",
    "                different different from z=0 along its propagation axis at t=0.\n",
    "        \"\"\"\n",
    "        return array([ self.amplitude(ν_i, x, z)\n",
    "                       * array([ cos(self.phase_angle(ν_i, x, z, TD))\n",
    "                                , - sin(self.phase_angle(ν_i, x, z, TD)) ])\n",
    "                      for ν_i in ν])\n",
    "    \n",
    "    \n",
    "    def amplitude(self, ν, x ,z):\n",
    "        \"\"\" Real spectral amplitude, includes additional terms from the approximation\n",
    "        \"\"\"\n",
    "        return ( self.spectrum(ν, x, z) * sqrt(self.funcLib.w0 / self.funcLib.w(z)) \n",
    "                 * exp( - ( x / self.funcLib.w(z) )**2 )\n",
    "                 * exp( - 2. * x * self.funcLib.beta_3(z) * 2. * sc.pi * (ν - self.funcLib.ν0) / self.funcLib.w(z) )\n",
    "               )\n",
    "    \n",
    "    \n",
    "    def phase_angle(self, ν, x , z, TD):\n",
    "        \"\"\" Spectral phase φ\n",
    "            \n",
    "            Analytic expectation after propagation of plane waves along z with angular dispersion \n",
    "            including initial time delay, spatial dispersion and group-delay dispersion.\n",
    "        \"\"\"\n",
    "        β_2 = self.funcLib.beta_2(x, z)\n",
    "        β_3 = self.funcLib.beta_3(z)\n",
    "        β_4 = self.funcLib.beta_4(z)\n",
    "        β_5 = self.funcLib.beta_5(z)\n",
    "        β_6 = self.funcLib.beta_6(x)\n",
    "        β_7 = self.funcLib.beta_7(z)\n",
    "        \n",
    "        γ_2 = self.funcLib.gamma_2(x, z)\n",
    "        γ_4_t_less = self.funcLib.gamma_4(t=0, x=x, z=z)\n",
    "        \n",
    "        return ( self.Ω0 * x**2 * self.funcLib.R_inv(z) / (2. * sc.c)\n",
    "            + self.Ω0 * z / sc.c\n",
    "            - .5 * arctan( z / self.zR )\n",
    "            + 0.25 * γ_2 * ( self.τ0 * 2. * sc.pi * (ν - self.ν0) )**2\n",
    "            - γ_4_t_less * ( self.τ0 * 2. * sc.pi * (ν - self.ν0) )\n",
    "            + 2. * sc.pi * (ν - self.ν0) * TD\n",
    "            + .5 * self.GDD_foc * ( 2. * sc.pi * (ν - self.ν0) )**2\n",
    "        )\n",
    "        \n",
    "    \n",
    "    def spectrum(self, ν, x, z):\n",
    "        \"\"\" Gaussian spectrum yielding a pulse with an FWHM duration of the intensity of\n",
    "                tau_FWHM,I = sqrt(2.*ln(2.)) * tau0\n",
    "            Additional terms of order (Ω-Ω)**2 in the real part of the field's exponent are added\n",
    "            here, too.\n",
    "        \"\"\"\n",
    "        norm = sqrt(sc.pi) * self.τ0\n",
    "        \n",
    "        γ_1 = self.funcLib.gamma_1(x, z)\n",
    "        if not ( γ_1 > 0 ):\n",
    "            print(\"Wotcher! Что ты делаешь? γ_1 > 0 ? γ_1 = %f\"%(γ_1))\n",
    "        \n",
    "        return norm * exp(- γ_1 * ( self.τ0 * sc.pi * (ν - self.ν0) )**2 )\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Implement real-valued time-space domain electric field"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (1) Derive relations of amplitude and phase of frequency-space domain field to the time-space domain field's real part."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "A periodic, real function $f(t)$, with a period of $T = (2n + 1)\\Delta t$ is trigonometrically interpolated by\n",
    "\n",
    "$$\n",
    "  f(t) \\approx p_n(t) = \\frac{1}{2n+1} \\left[\n",
    "    \\frac{a_0}{2} \n",
    "    + \\sum\\limits_{k=1}^{n} a_k \\cos(2\\pi\\nu_k t)\n",
    "      + b_k \\sin(2\\pi\\nu_k t)\n",
    "  \\right]\\,,\n",
    "$$\n",
    "where\n",
    "$$\n",
    "  \\nu_k = \\frac{k}{(2n+1)\\Delta t}\\,.\n",
    "$$\n",
    "\n",
    "The coefficients can be calculated from samples of the real space signal at times $t_\\ell = \\ell\\Delta t$ by\n",
    "$$\n",
    "\\begin{align}\n",
    "  a_k &= 2 \\sum\\limits_{\\ell=0}^{2n}f(t_\\ell)\\cos\\left( 2\\pi \\frac{k}{T} t_\\ell \\right) \\\\\n",
    "  b_k &= 2 \\sum\\limits_{\\ell=0}^{2n}f(t_\\ell)\\sin\\left( 2\\pi \\frac{k}{T} t_\\ell \\right)\\,.\n",
    "\\end{align}\n",
    "$$\n",
    "\n",
    "If the real space behavior of a signal is not known, but instead its complex spectrum given,\n",
    "$$\n",
    "  \\hat f (\\nu) = \\hat f_\\mathrm{amp} (\\nu) \\exp^{-\\imath \\varphi(\\nu)}\n",
    "$$\n",
    "the coefficients $(a_k, b_k)$ can be derived from the connection of the spectrum to the real space signal due to Fourier transformation on the one hand, and to its connection to the complex coefficients $c_k$ of the Fourier series of the real space signal on the other.\n",
    "\n",
    "That is, starting with the Fourier series\n",
    "$$\n",
    "\\begin{align}\n",
    "  f(t) &= \\frac{1}{2n+1}\\sum\\limits_{k=-\\infty}^{+\\infty} c_k e^{\\imath 2\\pi\\frac{k}{T}t}\n",
    "        \\approx \\frac{1}{2n+1}\\sum\\limits_{k=-n}^{+n} c_k e^{\\imath 2\\pi\\frac{k}{T}t} \\\\\n",
    "  c_k &= \\frac{1}{\\Delta t}\\int\\limits_{-T/2}^{T/2} f(t) e^{-\\imath 2\\pi\\frac{k}{T}t} dt\n",
    "       \\approx \\sum\\limits_{\\ell = 0}^{2n} f(t_\\ell) e^{-\\imath 2\\pi\\frac{k}{T}t_\\ell}\\,,\n",
    "\\end{align}\n",
    "$$\n",
    "to which the coefficients are connected by\n",
    "$$\n",
    "\\begin{align}\n",
    "  a_k &= c_k + c_{-k} \\\\\n",
    "  b_k &= i( c_k - c_{-k} )\\,.\n",
    "\\end{align}\n",
    "$$\n",
    "Given the spectrum $\\hat f(\\nu)$ of the signal, the coefficients $c_k$ can be deduced from the signal's Fourier transform\n",
    "$$\n",
    "  f(t) := \\int\\limits_{-\\infty}^{+\\infty} \\hat f(\\nu) e^{\\imath 2\\pi \\nu t} d\\nu\n",
    "$$\n",
    "which is connected to the complex coefficient by\n",
    "$$\n",
    "\\begin{align}\n",
    "  c_k &= \\frac{1}{\\Delta t}\\int\\limits_{-T/2}^{T/2}\n",
    "    \\int\\limits_{-\\infty}^{+\\infty} \\hat f(\\nu)\n",
    "    e^{\\imath 2\\pi( \\nu - \\frac{k}{T} )t} d\\nu dt \\\\\n",
    "  &= \\frac{1}{\\Delta t} \\int\\limits_{-\\infty}^{+\\infty} \\hat f(\\nu) \\delta\\left(\\nu - \\frac{k}{T}\\right) d\\nu \\\\\n",
    "  &= \\frac{1}{\\Delta t} \\hat f\\left( \\nu_k = \\frac{k}{T} \\right)\\,.\n",
    "\\end{align}\n",
    "$$\n",
    "Therefore,\n",
    "$$\n",
    "\\begin{align}\n",
    "  a_k = \\frac{1}{\\Delta t}\\left( \\hat f(\\nu_k) + \\hat f(\\nu_{-k}) \\right)\n",
    "    = 2 \\Re\\left\\lbrace \\hat f(\\nu_k) \\right\\rbrace / \\Delta t\n",
    "    = 2 \\hat f_\\mathrm{amp}(\\nu_k) \\cos\\varphi(\\nu_k) / \\Delta t \\\\\n",
    "  b_k = \\imath\\frac{1}{\\Delta t}\\left( \\hat f(\\nu_k) - \\hat f(\\nu_{-k}) ) \\right)\n",
    "    = - 2 \\Im\\left\\lbrace \\hat f(\\nu_k) \\right\\rbrace / \\Delta t\n",
    "    = 2 \\hat f_\\mathrm{amp}(\\nu_k) \\sin\\varphi(\\nu_k) / \\Delta t,\n",
    "\\end{align}\n",
    "$$\n",
    "provided that $\\hat f(\\nu_{-k}) = \\hat f^\\ast(\\nu_k)$ such that the coefficients are real.\n",
    "\n",
    "<span style=\"color:orange; font-weight:bolder\">\n",
    "    Note, assuming $\\hat f(\\nu_{-k}) = \\hat f^\\ast(\\nu_k)$ means to have two pulses,\n",
    "    one at $\\Omega = \\Omega_0$ and one at $\\Omega = -\\Omega_0$,\n",
    "    which effectively doubles the spectral field strength.\n",
    "    In order to compensate for that, the coefficients of the time domain field are divided by $2$.\n",
    "</span>\n",
    "\n",
    "That is,\n",
    "$$\n",
    "\\begin{align}\n",
    "  a_k = \\hat f_\\mathrm{amp}(\\nu_k) \\cos\\varphi(\\nu_k) / \\Delta t \\\\\n",
    "  b_k = \\hat f_\\mathrm{amp}(\\nu_k) \\sin\\varphi(\\nu_k) / \\Delta t,\n",
    "\\end{align}\n",
    "$$"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Given the complex spectrum of the field, it is also easy to infer the intensity envelope of the real space signal.\n",
    "Writing the $c_k$ as\n",
    "$$\n",
    "  c_{k,\\mathrm{Re}} + \\imath c_{k,\\mathrm{Im}} = \\frac{\\hat f_\\mathrm{amp}(\\nu_k)}{\\Delta t}\\left[ \\cos\\varphi(\\nu_k) - \\imath \\sin\\varphi(\\nu_k) \\right]\\,,\n",
    "$$\n",
    "the sum performed in the construction of $f(t)$ can be performed for the real and imaginary part seperately, yielding\n",
    "$$\n",
    "  f(t) = \\Re\\left\\lbrace f(t)\\right\\rbrace + \\imath \\Im\\left\\lbrace f(t)\\right\\rbrace = f_\\mathrm{Re}(t) + \\imath f_\\mathrm{Im}(t)\\,.\n",
    "$$\n",
    "The intensity envelope then simply is\n",
    "$$\n",
    "  \\lvert f(t) \\rvert^2 = f(t) \\cdot f^\\ast(t) = f_\\mathrm{Re}(t)^2 + f_\\mathrm{Im}(t)^2\\,.\n",
    "$$"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (2) Real-valued time-space electric field from complex-valued frequency-space field"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class RealSpaceLaserPulse(object):\n",
    "    \"\"\" \n",
    "        Elecric field of a laser pulse in time-space domain.\n",
    "        Computed from the frequency-space domain implementation by numerically Fourier transforming with a DFT.\n",
    "        See derivation above.\n",
    "    \"\"\"\n",
    "\n",
    "    def __init__(self,FourierSpaceLaserPulse, N_λ = 8., N_T = 14.):\n",
    "        \"\"\" Evaluate the sampling parameters for the real space field.\n",
    "                FourierSpaceLaserPulse - definition of the laser pulse in Fourier space\n",
    "                N_λ - number of samples per laser wavelength\n",
    "                N_T - length of time sampling interval in pulse durations\n",
    "            The latter two parameters define the length and spacing of the sampling interval in spectral domain\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.Δt = self.spectralLaser.λ_Laser / ( N_λ * sc.c )\n",
    "        \n",
    "        sampling_period = N_T * self.spectralLaser.τ0\n",
    "        self.n_dft = int( ( int( sampling_period / self.Δt ) - 1 ) / 2 )\n",
    "        print(\"Number of longitudinal samples =\", self.n_dft)\n",
    "        \n",
    "        self.ν = arange(0, self.n_dft + 1) / (2. * self.n_dft + 1.) / self.Δt\n",
    "        self.t = arange(-self.n_dft, self.n_dft+1, 1) * self.Δt ## theroretical range of times to look at\n",
    "\n",
    "        print(\"ν_max / ν_0 =\", self.ν[-1]/self.spectralLaser.ν0)\n",
    "        \n",
    "    \n",
    "    def __call__(self, x, z):\n",
    "        \"\"\" Return temporal evolution of electric field in real space\n",
    "            at position (x,z).\n",
    "            Returns an array with field values at times [-n_dft, n_dft]\n",
    "            where ( 2*n_dft + 1 ) * Δt = N_T * τ_Laser.\n",
    "        \"\"\"\n",
    "        # Multiplication of the coefficients with .5  here,\n",
    "        # because I actually have two pulses in frequency-space domain\n",
    "        # (at ν=ν_0 and ν=-ν_0)\n",
    "        # in order to comply with f(-ν) = f*(ν) which I need in order \n",
    "        # to obtain a purely real signal in time domain.\n",
    "        # Question: Is this equivalent to taking the real part of the complex\n",
    "        # signal I would obtain when transforming the complex frequency\n",
    "        # space pulse defined only for positive frequencies?\n",
    "        TD = - z / sc.c # Define time delay in order to locate the pulse's temporal origin at the position z where the field is evaluated.\n",
    "        a0 = self.spectralLaser(array([0.]), x, z, TD)[0][0] / self.Δt\n",
    "        c = self.spectralLaser(self.ν[1:], x, z, TD)\n",
    "        a = c[:, 0] / self.Δt\n",
    "        b = -c[:, 1] / self.Δt\n",
    "        \n",
    "        return array([ ( .5*a0 + sum( a * cos( 2. * sc.pi * self.ν[1:] * t_i )\n",
    "                                      + b * sin( 2. * sc.pi * self.ν[1:] * t_i )\n",
    "                                    )\n",
    "                       ) / (2.*self.n_dft + 1)\n",
    "                       for t_i in self.t\n",
    "                     ])\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (3) Real-valued time-space intensity envelope from complex-valued frequency-space field"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class RealSpaceIntensityEnvelope(object):\n",
    "    \"\"\" \n",
    "        Envelope representing the intensity contour of a laser pulse in time-space domain.\n",
    "        Computed from the frequency-space domain implementation by numerically Fourier transforming with a DFT.\n",
    "        See derivation above.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self,FourierSpaceLaserPulse, N_λ = 8., N_T = 14.):\n",
    "        \"\"\" Evaluate the sampling parameters for the real space intensity envelope.\n",
    "                FourierSpaceLaserPulse - definition of the laser pulse in Fourier space\n",
    "                N_λ - number of samples per laser wavelength\n",
    "                N_T - length of time sampling interval in pulse durations\n",
    "            The latter two parameters define the length and spacing of the sampling interval in spectral domain\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.Δt = self.spectralLaser.λ_Laser / ( N_λ * sc.c )\n",
    "        \n",
    "        sampling_period = N_T * self.spectralLaser.τ0\n",
    "        self.n_dft = int( ( int( sampling_period / self.Δt ) - 1 ) / 2 )\n",
    "        print(\"Number of longitudinal samples =\", self.n_dft)\n",
    "        \n",
    "        self.ν = arange(-self.n_dft, self.n_dft+1, 1) / (2. * self.n_dft + 1.) / self.Δt\n",
    "        self.t = arange(-self.n_dft, self.n_dft+1, 1) * self.Δt ## theroretical range of times to look at\n",
    "\n",
    "        print(\"ν_max / ν_0 =\", self.ν[-1]/self.spectralLaser.ν0)\n",
    "\n",
    "    \n",
    "    def __call__(self, x, z):\n",
    "        \"\"\" Return temporal evolution of the laser intensity envelope in time-space domain at position (x,z).\n",
    "            Returns an array with intensity envelope values at times [-n_dft, n_dft]\n",
    "            where ( 2*n_dft + 1 ) * Δt = N_T * τ_Laser.\n",
    "        \"\"\"\n",
    "        TD = - z / sc.c # Define time delay in order to locate the pulse's temporal origin at the position z where the field is evaluated.\n",
    "        c = self.spectralLaser(self.ν, x, z, TD) / self.Δt\n",
    "        \n",
    "        c_Real = c[:,0]\n",
    "        c_Imag = c[:,1]\n",
    "        \n",
    "        field_Real = array([ sum( c_Real * cos( 2. * sc.pi * self.ν * t_i )\n",
    "                                  - c_Imag  * sin( 2. * sc.pi * self.ν * t_i )\n",
    "                                ) for t_i in self.t ]) / (2.*self.n_dft + 1)\n",
    "        \n",
    "        field_Imag = array([ sum( c_Imag * cos( 2. * sc.pi * self.ν * t_i )\n",
    "                                  + c_Real  * sin( 2. * sc.pi * self.ν * t_i )\n",
    "                                ) for t_i in self.t ]) / (2.*self.n_dft + 1)\n",
    "        \n",
    "        intens_envelope = field_Real**2 + field_Imag**2\n",
    "        \n",
    "        return intens_envelope\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Analytic expressions for time-space domain properties of Gaussian laser pulses"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (1) Pulse-front tilt (full term)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class PulseFrontTiltAnalytically(object):\n",
    "    \"\"\" \n",
    "        Calculate the tangent of the pulse-front tilt angle at any position z along the propagation axis.\n",
    "        Compare eq. (18) in\n",
    "        K. Steiniger et al, \"Distortions in focusing laser pulses due to spatio-temporal couplings–An analytic description\".\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self,FourierSpaceLaserPulse):\n",
    "        \"\"\" Store the FourierSpaceLaserPulse in order to obtain the \n",
    "            laser's parameters.\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.funcLib = FourierSpaceFunctionLibrary(\n",
    "            self.spectralLaser.λ_Laser\n",
    "            , self.spectralLaser.τ0\n",
    "            , self.spectralLaser.w0\n",
    "            , self.spectralLaser.AD_foc\n",
    "            , self.spectralLaser.SD_foc\n",
    "            , self.spectralLaser.GDD_foc\n",
    "        )\n",
    "    \n",
    "    def __call__(self, z):\n",
    "        Ω0 = 2.*sc.pi*self.spectralLaser.ν0\n",
    "        init = - Ω0 * self.spectralLaser.AD_foc\n",
    "        \n",
    "        w = self.funcLib.w(z)\n",
    "        R_inv = self.funcLib.R_inv(z)\n",
    "        β_3 = self.funcLib.beta_3(z)\n",
    "        β_5 = self.funcLib.beta_5(z)\n",
    "        β_7 = self.funcLib.beta_7(z)\n",
    "        propagation_addition_1 = 2. * sc.c * β_5 * β_3 * Ω0 / w\n",
    "        propagation_addition_2 = 8. * sc.c * β_3 * (\n",
    "            β_7 - β_3**2 * β_5 * Ω0 - sc.c * self.spectralLaser.GDD_foc\n",
    "        ) / ( w * ( self.spectralLaser.τ0**2 + 4. * β_3**2 ) )\n",
    "        \n",
    "        return init + propagation_addition_1 + propagation_addition_2\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (2) Individual terms in pulse-front tilt expression for comparison later"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class PulseFrontTiltAnalyticallyTerm2(object):\n",
    "    \"\"\" Calculate the tangent of the pulse-front tilt angle at any position z along the propagation axis.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self,FourierSpaceLaserPulse):\n",
    "        \"\"\" Store the FourierSpaceLaserPulse in order to obtain the \n",
    "            laser's parameters.\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.funcLib = FourierSpaceFunctionLibrary(\n",
    "            self.spectralLaser.λ_Laser\n",
    "            , self.spectralLaser.τ0\n",
    "            , self.spectralLaser.w0\n",
    "            , self.spectralLaser.AD_foc\n",
    "            , self.spectralLaser.SD_foc\n",
    "            , self.spectralLaser.GDD_foc\n",
    "        )\n",
    "    \n",
    "    def __call__(self, z):\n",
    "        Ω0 = 2.*sc.pi*self.spectralLaser.ν0\n",
    "\n",
    "        w = self.funcLib.w(z)\n",
    "        R_inv = self.funcLib.R_inv(z)\n",
    "        β_3 = self.funcLib.beta_3(z)\n",
    "        β_5 = self.funcLib.beta_5(z)\n",
    "        β_7 = self.funcLib.beta_7(z)\n",
    "        propagation_addition_1 = 2. * sc.c * β_5 * β_3 * Ω0 / w\n",
    "        \n",
    "        return propagation_addition_1\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class PulseFrontTiltAnalyticallyTerm3(object):\n",
    "    \"\"\" Calculate the tangent of the pulse-front tilt angle at any position z along the propagation axis.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self,FourierSpaceLaserPulse):\n",
    "        \"\"\" Store the FourierSpaceLaserPulse in order to obtain the \n",
    "            laser's parameters.\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.funcLib = FourierSpaceFunctionLibrary(\n",
    "            self.spectralLaser.λ_Laser\n",
    "            , self.spectralLaser.τ0\n",
    "            , self.spectralLaser.w0\n",
    "            , self.spectralLaser.AD_foc\n",
    "            , self.spectralLaser.SD_foc\n",
    "            , self.spectralLaser.GDD_foc\n",
    "        )\n",
    "    \n",
    "    def __call__(self, z):\n",
    "        Ω0 = 2.*sc.pi*self.spectralLaser.ν0\n",
    "        \n",
    "        w = self.funcLib.w(z)\n",
    "        R_inv = self.funcLib.R_inv(z)\n",
    "        β_3 = self.funcLib.beta_3(z)\n",
    "        β_5 = self.funcLib.beta_5(z)\n",
    "        β_7 = self.funcLib.beta_7(z)\n",
    "        \n",
    "        propagation_addition_2 = 8. * sc.c * β_3 * (\n",
    "            β_7 - β_3**2 * β_5 * Ω0 - sc.c * self.spectralLaser.GDD_foc\n",
    "        ) / ( w * ( self.spectralLaser.τ0**2 + 4. * β_3**2 ) )\n",
    "        \n",
    "        return propagation_addition_2\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (3) Pulse front"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class PulseFront(object):\n",
    "    \"\"\" Calculate the time offset t_0 of the real space field, which depends on the transverse coordinate x,\n",
    "        at any position z along the propagation axis.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self,FourierSpaceLaserPulse):\n",
    "        \"\"\" Store the FourierSpaceLaserPulse in order to obtain the \n",
    "            laser's parameters.\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.funcLib = FourierSpaceFunctionLibrary(\n",
    "            self.spectralLaser.λ_Laser\n",
    "            , self.spectralLaser.τ0\n",
    "            , self.spectralLaser.w0\n",
    "            , self.spectralLaser.AD_foc\n",
    "            , self.spectralLaser.SD_foc\n",
    "            , self.spectralLaser.GDD_foc\n",
    "        )\n",
    "    \n",
    "    def __call__(self, x, z):\n",
    "        #init = - self.Ω0 * self.spectralLaser.AD_foc\n",
    "        \n",
    "        Ω0 = self.spectralLaser.Ω0\n",
    "        τ0 = self.spectralLaser.τ0\n",
    "        w = self.funcLib.w(z)\n",
    "        R_inv = self.funcLib.R_inv(z)\n",
    "        zR = self.spectralLaser.zR\n",
    "        β_2 = self.funcLib.beta_2(x, z = 0) # Neglect the z dependence in order to draw te pulse-front always through the center of the pulse.\n",
    "        β_3 = self.funcLib.beta_3(z)\n",
    "        β_4 = self.funcLib.beta_4(z)\n",
    "        β_5 = self.funcLib.beta_5(z)\n",
    "        β_6 = self.funcLib.beta_6(x)\n",
    "        β_7 = self.funcLib.beta_7(z)\n",
    "        \n",
    "        second = β_5 * x**2 / w**2\n",
    "        \n",
    "        third = 2. * (x / w) * β_3 * Ω0 * β_5\n",
    "        \n",
    "        GDD = (\n",
    "            self.spectralLaser.GDD_foc\n",
    "            + 4. * β_3 * β_5 * x / w \n",
    "            + 2. * ( 2. * β_4 * x / w + β_3**2 ) * Ω0 * β_5\n",
    "            - 2. * (β_6 + β_7)\n",
    "        )\n",
    "        \n",
    "        tau2 = τ0**2 + 8. * β_4 * x / w + 4. * β_3**2\n",
    "        \n",
    "        fourth = - 4. * ( x * β_3 / w) * GDD / tau2\n",
    "        \n",
    "        t0 = β_2 + second + third + fourth\n",
    "        \n",
    "        γ_1 = self.funcLib.gamma_1(x, z)\n",
    "        if not ( γ_1 > 0 ):\n",
    "            print(\"Wotcher! Что ты делаешь? γ_1 > 0 ? γ_1 = %f\"%(γ_1))\n",
    "        \n",
    "        return t0\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### (4) Pulse duration"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "class PulseDurationAnalytically(object):\n",
    "    \"\"\" Calculate the pulse duration as the FWHM of the real space intensity\n",
    "        at any position z along the propagation axis.\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self,FourierSpaceLaserPulse):\n",
    "        \"\"\" Store the FourierSpaceLaserPulse in order to obtain the \n",
    "            laser's parameters.\n",
    "        \"\"\"\n",
    "        self.spectralLaser = FourierSpaceLaserPulse\n",
    "        self.funcLib = FourierSpaceFunctionLibrary(\n",
    "            self.spectralLaser.λ_Laser\n",
    "            , self.spectralLaser.τ0\n",
    "            , self.spectralLaser.w0\n",
    "            , self.spectralLaser.AD_foc\n",
    "            , self.spectralLaser.SD_foc\n",
    "            , self.spectralLaser.GDD_foc\n",
    "        )\n",
    "    \n",
    "    def __call__(self, z):\n",
    "        \"\"\"\n",
    "        \"\"\"\n",
    "        τ0 = self.spectralLaser.τ0\n",
    "        γ_1_center = self.funcLib.gamma_1(x=0, z=z)\n",
    "        γ_2_center = self.funcLib.gamma_2(x=0, z=z)\n",
    "        return τ0 * sqrt( γ_1_center + γ_2_center**2 / γ_1_center )\n",
    "        \n",
    "        \n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Implementation of library-class which computes dispersion parameters in the focus of an off-axis parabola from parameters before focusing at the parabola\n",
    "\n",
    "Compare with sec. \"Deriving pulse dispersion in the focus of an off-axis parabola\" in K. Steiniger et al, \"Distortions in focusing laser pulses due to spatio-temporal couplings–An analytic description\"."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "from scipy.misc import derivative as scipy_misc_derivative \n",
    "\n",
    "\n",
    "class DispersionInFocus(object):\n",
    "    \"\"\" Calculate disperison properties (SD, AD, GDD, TOD, ...) of a laser pulse in the\n",
    "        focus of a parabola from its values before the mirror.\n",
    "        The in-focus values are available by accessing the corresponding member variables directly\n",
    "    \"\"\"\n",
    "    \n",
    "    def __init__(self\n",
    "                 , wavelength = 800.e-9 # wavelength of the central laser frequency\n",
    "                                        # unit: m\n",
    "                 , parent_focal_distance = 2.5 # distance of the focal point from the vertex of the parabola\n",
    "                                               # unit: m\n",
    "                 , deflection_angle = 45. * sc.pi / 180. # deflection angle of the central frequency for\n",
    "                                                         # incidence parallel to the axis of the parabola\n",
    "                                                         # unit: rad\n",
    "                 , angular_dispersion_at_mirror = 0. # dθ/dΩ, unit: rad*s, z.B. tan(α_tilt)/(2.*sc.pi*ν0)\n",
    "                                                     # = (dθ/dλ)*(dλ/dΩ) = (dθ/dλ) * (-1) * λ0^2 / (2πc)\n",
    "                 , spatial_dispersion_at_mirror = 0. # dx/dΩ, unit: m*s, \n",
    "                                                     # = (dx/dλ)*(dλ/dΩ) = (dx/dλ) * (-1) * λ0^2 / (2πc)\n",
    "                 , group_delay_dispersion_at_mirror = 0. # (d^2φ/dΩ^2), unit: s^2\n",
    "                                                         # = (d^2φ/dν^2)*(dν/dΩ)^2 = (d^2φ/dν^2) / (2π)^2\n",
    "                 , third_order_dispersion_at_mirror = 0. # (d^3φ/dΩ^3), unit: s^3\n",
    "                                                         # = (d^3φ/dν^3)*(dν/dΩ)^3 = (d^3φ/dν^3) / (2π)^3\n",
    "                ):\n",
    "                \n",
    "        self.λ0 = wavelength\n",
    "        self.ν0 = sc.c / self.λ0\n",
    "        self.f = parent_focal_distance\n",
    "        self.ψ_defl0 = deflection_angle\n",
    "        self.f_eff0 = parent_focal_distance / cos( .5 * self.ψ_defl0 )**2 # effective focal distance of parabola\n",
    "                                                                          # for central frequency\n",
    "        self.ξ_s0 = 2. * self.f * tan( .5 * self.ψ_defl0 ) # point of incidence of central frequency on parabola\n",
    "        self.AD_mirr = angular_dispersion_at_mirror\n",
    "        self.SD_mirr = spatial_dispersion_at_mirror\n",
    "        self.GDD_mirr = group_delay_dispersion_at_mirror\n",
    "        self.TOD_mirr = third_order_dispersion_at_mirror\n",
    "        self.h = 1.e-4*self.ν0 # step size for finite difference approximation of derivatives\n",
    "        self.dνdΩ = 1. / ( 2. * sc.pi ) # inner derivative with respect to Ω of functions\n",
    "                                        # defined with respect to ν\n",
    "        self.SD_foc = self.calc_SD_in_focus()\n",
    "        self.AD_foc = self.calc_AD_in_focus()\n",
    "        self.GDD_foc = self.calc_GDD_in_focus()\n",
    "        self.TOD_foc = self.calc_TOD_in_focus()\n",
    "    \n",
    "    \n",
    "    def __call__(self):\n",
    "        print(\"Parent focal distance = %.1fmm; Effective focal distance = %.1fmm\"%(self.f*1.e3, self.f_eff0*1.e3))\n",
    "        CONV_MS_TO_MpM = 1. / ( -1*self.λ0**2 / (2.*sc.pi*sc.c) )\n",
    "        print(\"SD @ mirror = %.3emm/nm; SD @ focus = %.3emm/nm\"%(self.SD_mirr * CONV_MS_TO_MpM * 1.e-6, self.SD_foc * CONV_MS_TO_MpM * 1.e-6))\n",
    "        CONV_RADS_TO_RADpM = CONV_MS_TO_MpM\n",
    "        print(\"AD @ mirror = %.3eµrad/nm; AD @ focus = %.3eµrad/nm\"%(\n",
    "            self.AD_mirr * CONV_RADS_TO_RADpM * 1.e-3\n",
    "            , self.AD_foc * CONV_RADS_TO_RADpM * 1.e-3\n",
    "        ))\n",
    "        print(\"PFT @ mirror = %.3edeg; PFT @ focus = %.3edeg\"%(\n",
    "            arctan(2.*sc.pi*self.ν0 * self.AD_mirr ) * 180./ sc.pi\n",
    "            , arctan(2.*sc.pi*self.ν0 * self.AD_foc ) * 180./ sc.pi\n",
    "        ))\n",
    "        print(\"GDD @ mirror = %.3efs^2; GDD @ focus = %.3efs^2\"%(self.GDD_mirr*(1.e15)**2, self.GDD_foc*(1.e15)**2))\n",
    "        print(\"TOD @ mirror = %.3efs^3; TOD @ focus = %.3efs^3\"%(self.TOD_mirr*(1.e15)**3, self.TOD_foc*(1.e15)**3))\n",
    "        \n",
    "    def calc_AD_in_focus(self):\n",
    "        \"\"\" Angular dispersion (AD_{foc,coupl}) in focus obtained from first derivative of frequency dependent\n",
    "            angle enclosed by the propagation direction of frequency ν and the central frequency ν0\n",
    "            \n",
    "            Eq. (22)\n",
    "        \"\"\"\n",
    "        return - self.SD_mirr / self.f_eff0 - self.AD_mirr\n",
    "\n",
    "    def calc_SD_in_focus(self):\n",
    "        \"\"\" Sptial dispersion (SD_{foc,coupl}) in focus obtained from first derivative of\n",
    "            distance between spatial distribution centers of frequency ν\n",
    "            and the central frequency ν0\n",
    "            (Derivative performed analytically. Only result here.)\n",
    "            \n",
    "            Eq. (23)\n",
    "        \"\"\"\n",
    "        return self.f_eff0 * self.AD_mirr\n",
    "    \n",
    "    def calc_GDD_in_focus(self):\n",
    "        \"\"\" Group delay dispersion (GDD_{foc,coupl}) in the focus obtained from second derivative of the phase\n",
    "        \n",
    "            Eq. (24)\n",
    "        \"\"\"\n",
    "        return -2. * sc.pi * self.ν0 * (\n",
    "            self.f_eff0 * self.AD_mirr**2 + 2. * self.SD_mirr * self.AD_mirr\n",
    "        ) / sc.c + self.GDD_mirr\n",
    "    \n",
    "    def calc_TOD_in_focus(self):\n",
    "        \"\"\" Third order dispersion (TOD_{foc,coupl}) in focus obtained from third derivative of the phase\n",
    "        \n",
    "            Eq. (25)\n",
    "        \"\"\"\n",
    "        return 3. * self.AD_mirr * (\n",
    "            2. * sc.pi * self.ν0 * self.ξ_s0 * self.SD_mirr * self.AD_mirr / self.f\n",
    "            - self.f_eff0 * self.AD_mirr\n",
    "            - 2. * self.SD_mirr\n",
    "        ) / sc.c + self.TOD_mirr\n",
    "    \n",
    "    def f_eff(self, ν):\n",
    "        \"\"\" Frequency dependent effective focal distance.\n",
    "        \"\"\"\n",
    "        return self.ξ_s(ν) / sin( self.ψ_defl(ν) )\n",
    "    \n",
    "    def ψ_defl(self, ν):\n",
    "        \"\"\" Frequency dependent deflection angle at parabola\n",
    "        \"\"\"\n",
    "        return 2. * arctan2( self.ξ_s(ν), 2.*self.f )\n",
    "    \n",
    "    def ξ_s(self, ν):\n",
    "        \"\"\" x coordinate of the point where incident ray of frequency ν\n",
    "            intersects with the parabola surface\n",
    "        \"\"\"\n",
    "        p = (self.ξ_s0 - self.x_0(ν))\n",
    "        q = 0.25 * self.ξ_s0**2 / self.f\n",
    "        return p + (q - 0.25*p**2/self.f) * self.θ_0(ν)\n",
    "    \n",
    "    def θ_0(self, ν):\n",
    "        \"\"\" Angle enclosed by the propagation direction of frequency ν\n",
    "            and the central frequency ν0 before deflection at the parabola\n",
    "        \"\"\"\n",
    "        return self.AD_mirr * 2. * sc.pi * (ν - self.ν0)\n",
    "    \n",
    "    def x_0(self, ν):\n",
    "        \"\"\" Distance between the spatial distribution center of frequency ν\n",
    "            and the central frequency ν0 before deflection at the parabola\n",
    "        \"\"\"\n",
    "        return self.SD_mirr * 2. * sc.pi * ( ν - self.ν0 )\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Plot intensity map and pulse-front for Gauss with tilt and dispersion values calculated from values before focusing mirror, Fig. 6."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true
   },
   "source": [
    "## (1) Input (dispersion) values (before off-axis parabola)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "# central wavelength of laser\n",
    "wavel = 800.e-9\n",
    "\n",
    "# pulse duration\n",
    "pulse_dur = 30.e-15 / 1.17741\n",
    "\n",
    "# deflection angle at parabola\n",
    "ψ_defl0 = 90. * sc.pi / 180.\n",
    "\n",
    "# f/# of off-axis parabola\n",
    "f_num = 2.5\n",
    "\n",
    "# beam size at parabola\n",
    "D = 100.e-3\n",
    "\n",
    "# Conversion factor from derivative with respect to wavelength λ\n",
    "# to derivative with respect to frequency Ω\n",
    "dλdΩ = (-1.) * wavel**2 / (2.*sc.pi*sc.c)\n",
    "\n",
    "# angular dispersion at parabola\n",
    "# as Δθ/ΔΩ from Δθ/Δλ in µrad/nm\n",
    "AD_in = -1. * 1.e-6 / 1.e-9 * dλdΩ\n",
    "\n",
    "# spatial dispersion at parabola\n",
    "# as Δx/ΔΩ from Δx/Δλ in mm/nm\n",
    "SD_in = -0. * 1.e-3 / 1.e-9 * dλdΩ\n",
    "\n",
    "# group delay dispersion at parabola\n",
    "# as Δ^2φ/ΔΩ^2 from Δ^2φ/ΔΩ^2 in (fs)^2\n",
    "GDD_in = 0. * (1.e-15)**2\n",
    "\n",
    "# third order dispersion at parabola\n",
    "# as Δ^3φ/ΔΩ^3 from Δ^3φ/ΔΩ^3 in (fs)^3\n",
    "TOD_in = 0. * (1.e-15)**3"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true
   },
   "source": [
    "## (2) Compute disperson in focus"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "LaserDispersionInFocus = DispersionInFocus(\n",
    "    wavelength = wavel\n",
    "    , parent_focal_distance = f_num * D *cos( .5 * ψ_defl0 )**2 # assuming f/# is for off-axis parabola,\n",
    "                                                                # i.e. f/# * D is the effective focal distance\n",
    "    , deflection_angle = ψ_defl0\n",
    "    , angular_dispersion_at_mirror = AD_in\n",
    "    , spatial_dispersion_at_mirror = SD_in\n",
    "    , group_delay_dispersion_at_mirror = GDD_in\n",
    "    , third_order_dispersion_at_mirror = TOD_in\n",
    ")\n",
    "\n",
    "LaserDispersionInFocus()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (3) Check whether requirements for neglecting TOD are fulfilled"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "print(\"τ_0 = %e\"%(pulse_dur))\n",
    "FuncLib = FourierSpaceFunctionLibrary(\n",
    "    wavelength = wavel\n",
    "    , fourier_limited_pulse_duration = pulse_dur\n",
    "    , fourier_limited_width_in_focus = focal_width\n",
    "    , angular_dispersion_in_focus = LaserDispersionInFocus.AD_foc\n",
    "    , spatial_dispersion_in_focus = 0. #LaserDispersionInFocus.SD_foc\n",
    "    , group_delay_dispersion_in_focus = 0. #LaserDispersionInFocus.GDD_foc\n",
    ")\n",
    "z_check = FuncLib.zR\n",
    "w_check = FuncLib.w(z_check)\n",
    "β_3 = FuncLib.beta_3(z_check)\n",
    "β_4 = FuncLib.beta_4(z_check)\n",
    "β_5 = FuncLib.beta_5(z_check)\n",
    "δ_1 = -z_check*LaserDispersionInFocus.AD_foc**3/(6.*w_check)\n",
    "δ_2 = z_check*LaserDispersionInFocus.AD_foc**2/sc.c\n",
    "TOD_check = 6.*β_3**2*β_5 + 12.*FuncLib.Ω0*β_3*β_4*β_5 - 6.*δ_2\n",
    "req1 = 128.*β_3*β_4/pulse_dur**3\n",
    "print(\"128 β_3 β_4 / τ_0^3 = %e << 1?\"%(req1))\n",
    "req2 = 11.*TOD_check/pulse_dur**3\n",
    "print(\"11 TOD(z) / τ_0^3 = %e << 1?\"%(req2))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (4) Compute time-space domain properties"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "focal_width = LaserDispersionInFocus.f_eff0 * wavel / D\n",
    "print(\"focal width = %.2fµm\"%(focal_width * 1.e6))\n",
    "\"\"\"\"\"\"\n",
    "LaserFourierSpace = GaussPulseFourierSpace(\n",
    "    wavelength = wavel\n",
    "    , fourier_limited_pulse_duration = pulse_dur\n",
    "    , fourier_limited_width_in_focus = focal_width\n",
    "    , angular_dispersion_in_focus = LaserDispersionInFocus.AD_foc\n",
    "    , spatial_dispersion_in_focus = LaserDispersionInFocus.SD_foc\n",
    "    , group_delay_dispersion_in_focus = 0. #LaserDispersionInFocus.GDD_foc\n",
    ")\n",
    "\n",
    "LaserRealSpace = RealSpaceLaserPulse(LaserFourierSpace, N_λ = 6, N_T = 14)\n",
    "LaserIntens = RealSpaceIntensityEnvelope(LaserFourierSpace, N_λ = 6, N_T = 14)\n",
    "LaserTanψ_tilt = PulseFrontTiltAnalytically(LaserFourierSpace)\n",
    "LaserPulseFront = PulseFront(LaserFourierSpace)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "print(\"SD_foc * ( Ω_egde - Ω_0 ) / w_0  @ focus = %.2f\"%( LaserTanψ_tilt.funcLib.beta_3(0.) / pulse_dur ))\n",
    "print(\"T / τ_0 @ focus = %.2f\"%( sqrt( LaserTanψ_tilt.funcLib.gamma_1(0., 0.) ) ))\n",
    "print(\"Expected ratio of intensity amplitude to undisturbed pulse = %.2f\"%( ( LaserTanψ_tilt.funcLib.gamma_1(0., 0.)**2 + LaserTanψ_tilt.funcLib.gamma_2(0., 0.)**2 )**(-0.5) ))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (5) Set space domain interval of plotting region"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "x = sc.c * LaserFourierSpace.τ0 * linspace(-2., 2., 129, endpoint=True) #LaserFourierSpace.w0 * linspace(-5., 5., 129, endpoint=True)\n",
    "z = LaserFourierSpace.zR * linspace(-5., 1., 7, endpoint=True)\n",
    "\n",
    "t_norm = LaserFourierSpace.τ0\n",
    "x_norm = sc.c * LaserFourierSpace.τ0\n",
    "\n",
    "print(\"z_R = %.2fµm\"%(LaserFourierSpace.zR*1.e6))\n",
    "print(\"z ϵ [%.2f, %.2f]z_R\"%(z[0]/LaserFourierSpace.zR, z[-1]/LaserFourierSpace.zR))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (6) Plot intensity map with pulse-front tilt calculation and pulse-front outline"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import functools\n",
    "from scipy.optimize import curve_fit\n",
    "\n",
    "def gauss(t, sig, amp, mean):\n",
    "    return amp * exp(-.5*(t - mean)**2/sig**2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "i=0\n",
    "for z_i in z:\n",
    "    \n",
    "    intens2D_t = array([ LaserIntens(x_i, z_i) for x_i in x ])\n",
    "    \n",
    "    ############################################################\n",
    "    ## Calculate PFT\n",
    "    ############################################################\n",
    "    ## By deducing arrival time of intensity maximum for two different x positions,\n",
    "    ## where the x positions are equally spaced to the left and right from the pulse's center,\n",
    "    ## and compute tangent of PFT angle w/ central finite difference.\n",
    "    ## Note, this is an updated algorithm compared to the one used in the manuscript,\n",
    "    ## where the finite difference used values in the center (x_center) and off-center (x_out).\n",
    "    Δx = int(1.*LaserFourierSpace.w0 / abs(x[1]-x[0]))\n",
    "    x_center = int(.5*len(x-1))\n",
    "    x_left = x_center - Δx\n",
    "    x_right = x_center + Δx\n",
    "    max_center = argmax(intens2D_t[x_center])\n",
    "    max_left = argmax(intens2D_t[x_left])\n",
    "    max_right = argmax(intens2D_t[x_right])\n",
    "    \n",
    "    tanψ_tilt_num = (\n",
    "        (sc.c*LaserIntens.t[max_left] - sc.c*LaserIntens.t[max_right])\n",
    "        / (x[x_left] - x[x_right])\n",
    "    )\n",
    "    ψ_tilt_num = arctan(tanψ_tilt_num)\n",
    "    \n",
    "    print(\"pulse-front tilt (numerically) = arctan( (%.2f - %.2f  ) / ( %.2f - %.2f  ) ) = %.2f deg\"%(\n",
    "        LaserIntens.t[max_left] / t_norm\n",
    "        , LaserIntens.t[max_right] / t_norm\n",
    "        , x[x_left] / x_norm\n",
    "        , x[x_right] / x_norm\n",
    "        , ψ_tilt_num * 180. / pi\n",
    "    ))\n",
    "    \n",
    "    ####################\n",
    "    ## Additional sample to quantify uncertainty on PFT value\n",
    "    ####################\n",
    "    Δx_sample = int(.5*Δx)\n",
    "    x_left_sample = x_center - Δx_sample\n",
    "    x_right_sample = x_center + Δx_sample\n",
    "    max_left_sample = argmax(intens2D_t[x_left_sample])\n",
    "    max_right_sample = argmax(intens2D_t[x_right_sample])\n",
    "    \n",
    "    tanψ_tilt_num_sample = (\n",
    "        (sc.c*LaserIntens.t[max_left_sample] - sc.c*LaserIntens.t[max_right_sample])\n",
    "        / (x[x_left_sample] - x[x_right_sample])\n",
    "    )\n",
    "    ψ_tilt_num_sample = arctan(tanψ_tilt_num_sample)\n",
    "    print(\"pulse-front tilt (numerically) sample = %.2f deg\"%(ψ_tilt_num_sample * 180. / pi))\n",
    "    \n",
    "    ####################\n",
    "    ## Estimate uncertainty due to finite sampling on computed pulse-front tilt\n",
    "    ####################\n",
    "    ## By adding one cell to each of the determined arrival times in the numerical tangent calculation\n",
    "    ## For the last z position.\n",
    "    tanψ_tilt_num_error_max = (\n",
    "        (sc.c*LaserIntens.t[max_left - 1] - sc.c*LaserIntens.t[max_right + 1])\n",
    "        / (x[x_left] - x[x_right])\n",
    "    )\n",
    "    ψ_tilt_num_error_max = arctan(tanψ_tilt_num_error_max)\n",
    "    print(\"Maximum pulse-front tilt (numerically) = %.2f deg\"%(ψ_tilt_num_error_max * 180. / pi))\n",
    "\n",
    "    tanψ_tilt_num_error_min = (\n",
    "        (sc.c*LaserIntens.t[max_left + 1] - sc.c*LaserIntens.t[max_right - 1])\n",
    "        / (x[x_left] - x[x_right])\n",
    "    )\n",
    "    ψ_tilt_num_error_min = arctan(tanψ_tilt_num_error_min)\n",
    "    print(\"Minimum pulse-front tilt (numerically) = %.2f deg\"%(ψ_tilt_num_error_min * 180. / pi))\n",
    "\n",
    "    print(\"Estimated uncertainty on pulse-front tilt angle = %f°\"%(.5 * (ψ_tilt_num_error_max - ψ_tilt_num_error_min) * 180. / pi))\n",
    "    \n",
    "    ############################################################\n",
    "    ## Calculate pulse duration\n",
    "    ############################################################\n",
    "    intensGauss = functools.partial(gauss, amp = intens2D_t[x_center][max_center], mean=LaserIntens.t[max_center])\n",
    "    intensGauss_popt, intensGauss_pcov = curve_fit(intensGauss, LaserIntens.t / LaserFourierSpace.τ0, intens2D_t[x_center], p0 = 1.)\n",
    "    τ_meas_τ_0 = 2. * intensGauss_popt[0] # measured duration of the field from the standard deviation of the intensity distribution, unit: τ_0\n",
    "    \n",
    "    ############################################################\n",
    "    ## Plot\n",
    "    ############################################################\n",
    "    figure(figsize=(16,6))\n",
    "    \n",
    "    print(\"pulse-front tilt (ana) = %.2f deg\"%( arctan(LaserTanψ_tilt(z_i)) * 180. / sc.pi ))\n",
    "    \n",
    "    subplot(1,2,1)\n",
    "    \n",
    "    imshow(intens2D_t * sqrt( LaserTanψ_tilt.funcLib.gamma_1(0., 0.)**2 + LaserTanψ_tilt.funcLib.gamma_2(0., 0.)**2  )\n",
    "           , cmap = mpl.cm.Reds\n",
    "           , origin='lower'\n",
    "           , interpolation='none'\n",
    "           , aspect='equal'\n",
    "           , extent=(LaserIntens.t[0] / t_norm\n",
    "                     , LaserIntens.t[-1] / t_norm\n",
    "                     , x[0] / x_norm\n",
    "                     , x[-1] / x_norm\n",
    "                    )\n",
    "           , vmin = 0., vmax = 1.\n",
    "           #, norm=matplotlib.colors.LogNorm(vmin=1.e-0)\n",
    "          )\n",
    "\n",
    "    #colorbar(ticks = linspace(0., 1., 6, endpoint=True), format='%.1f')\n",
    "    \n",
    "    ####################\n",
    "    ## Plot theoretical pulse-front contour.\n",
    "    ####################\n",
    "    ## Since I plot t on the x axis, and x ond the y axis, the slope of the pulse-front is 90°-ψ_tilt.\n",
    "    ## Here I use tan( 90°-ψ_tilt ) = 1 / tan( ψ_tilt )\n",
    "    plot(LaserTanψ_tilt(z_i) * x / sc.c / t_norm, x / x_norm, label=r\"analytic PFT = %.2f°\"%( arctan(LaserTanψ_tilt(z_i)) * 180. / sc.pi ) )\n",
    "    plot(tanψ_tilt_num * x / sc.c / t_norm, x / x_norm, '--', label=\"measured PFT = %.1f°\"%(ψ_tilt_num * 180. / pi))\n",
    "    #plot(array([LaserPulseFront(x_i, z_i) for x_i in x ]) / t_norm, x / x_norm, label=\"analytic front\" )\n",
    "    print(\"Difference of pulse maximum arrival time in cells at x=x_right between the analytic and numeric PFT angle = %.1f\"%(\n",
    "        (LaserTanψ_tilt(z_i) - tanψ_tilt_num) * x[x_right] / sc.c / LaserIntens.Δt\n",
    "    ))\n",
    "    \n",
    "    text(-3.5, -1.5, r\"measured duration = %.1f$\\tau_0$\"%( τ_meas_τ_0 ), fontsize=14)\n",
    "    \n",
    "    xlabel(r\"$t$ [$\\tau_0$]\")\n",
    "    ylabel(r\"$x$ [$c\\tau_0$]\")\n",
    "    \n",
    "    xlim(-4., 4.)\n",
    "    #ylim( -.5, .5 )\n",
    "    \n",
    "    title(\"Intensity @ z = %.2f z_R\"%(z_i /  LaserFourierSpace.zR))\n",
    "    \n",
    "    grid()\n",
    "    \n",
    "    legend()\n",
    "        \n",
    "    show()\n",
    "    #savefig(\"pics/DRACO_intens_ADmir-%.1fmurad-nm_%03i-%.1f-zR.png\"%(AD_in / dλdΩ * 1.e6 / 1.e9 ,i, z_i /  LaserFourierSpace.zR), dpi=100, transparent=False, bbox_inches='tight')\n",
    "\n",
    "    close()\n",
    "    i+=1\n",
    "\n",
    "del i"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Plot scaling of pulse-front tilt and pulse duration for a range of AD values and possibly SD at mirror, Figs. 4 and 5"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (1) Input (dispersion) values (before off-axis parabola)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# central wavelength of laser\n",
    "wavel = 800.e-9\n",
    "\n",
    "# pulse duration\n",
    "pulse_dur = 30.e-15 / 1.17741\n",
    "\n",
    "# deflection angle at parabola\n",
    "ψ_defl0 = 90. * sc.pi / 180. # 45. * sc.pi / 180.\n",
    "\n",
    "# f/# of off-axis parabola\n",
    "f_num = 2.5 # Fig. 4\n",
    "f_num = 20. # Fig. 5\n",
    "\n",
    "# beam size at parabola\n",
    "D = 100.e-3\n",
    "\n",
    "# effective focal distance\n",
    "f_effective = f_num * D\n",
    "print(\"effective focal distance of off-axis parabola = %.1fmm\"%(f_effective*1.e3))\n",
    "\n",
    "focal_width = f_effective * wavel / D\n",
    "print(\"focal width FWHM = %.2fµm\"%(focal_width * 1.17741 * 1.e6))\n",
    "\n",
    "Rayleigh_length = sc.pi*focal_width**2 / wavel\n",
    "print(\"Rayleigh length z_R = %.1fµm\"%( Rayleigh_length * 1.e6 ))\n",
    "\n",
    "# Conversion factor from derivative with respect to wavelength λ\n",
    "# to derivative with respect to frequency Ω\n",
    "dλdΩ = (-1.) * wavel**2 / (2.*sc.pi*sc.c)\n",
    "\n",
    "# angular dispersion at mirror\n",
    "# as Δθ/ΔΩ from Δθ/Δλ in µrad/nm\n",
    "AD_mir = array([5.e-3, 1.e-2, 2.5e-2, 5.e-2, 1.e-1, 2.5e-1, 5.e-1, 1.], dtype=longdouble) * 1.e-6 / 1.e-9 * dλdΩ\n",
    "\n",
    "# spatial dispersion at mirror\n",
    "# as Δx/ΔΩ from Δx/Δλ in mm/nm\n",
    "SD_mir = -zeros(len(AD_mir))\n",
    "print(\"Spatial dispersion at the mirror [µm/nm] = \", SD_mir / dλdΩ * 1.e6 / 1.e9)\n",
    "\n",
    "# angular and spatial dispersion in focus\n",
    "AD_foc = zeros(len(AD_mir), dtype=longdouble)\n",
    "SD_foc = zeros(len(AD_foc))\n",
    "GDD_foc = zeros(len(AD_foc))\n",
    "TOD_foc = zeros(len(AD_foc))\n",
    "for i, AD_mir_i in enumerate(AD_mir):\n",
    "    Dispersions = DispersionInFocus(\n",
    "        wavelength = wavel\n",
    "        , parent_focal_distance = f_effective *cos( .5 * ψ_defl0 )**2 \n",
    "        , deflection_angle = ψ_defl0\n",
    "        , angular_dispersion_at_mirror = AD_mir_i\n",
    "        , spatial_dispersion_at_mirror = SD_mir[i]\n",
    "        , group_delay_dispersion_at_mirror = 0.\n",
    "        , third_order_dispersion_at_mirror = 0.\n",
    "    )\n",
    "    AD_foc[i] = Dispersions.AD_foc\n",
    "    SD_foc[i] = Dispersions.SD_foc\n",
    "    GDD_foc[i] = Dispersions.GDD_foc\n",
    "    TOD_foc[i] = Dispersions.TOD_foc\n",
    "print(\"Angular dispersion in focus from AD and SD at mirror [µrad/nm] = \", AD_foc / dλdΩ * 1.e6 / 1.e9)\n",
    "print(\"Pulse-front tilt in the focus [°] =\",  arctan( -2.*sc.pi * sc.c / wavel * AD_foc ) * 180. / sc.pi )\n",
    "print(\"Spatial dispersion in focus from AD and SD at mirror [µm/nm] = \", SD_foc / dλdΩ * 1.e6 / 1.e9)\n",
    "print(\"Group delay dispersion in focus from AD and SD at mirror [(fs)**2] = \", GDD_foc * (1.e15)**2)\n",
    "print(\"Third order dispersion in focus from AD and SD at mirror [(fs)**3] = \", TOD_foc * (1.e15)**3)\n",
    "\n",
    "print(\"Pulse elongation in the focus due to GDD [1/τ_0] =\", sqrt( 1 + 4. * GDD_foc**2 / pulse_dur**4 ))\n",
    "print(\"I NEGLECT GDD_FOC AND TOD_FOC BECAUSE THEY ARE SO LITTLE.\")\n",
    "\n",
    "# group delay dispersion in focus\n",
    "# as Δ^2φ/ΔΩ^2 from Δ^2φ/ΔΩ^2 in (fs)^2\n",
    "GDD_foc = 0. * (1.e-15)**2 # well collimated beam: -Ω0 * θ'^2 * z/c\n",
    "\n",
    "# third order dispersion in focus\n",
    "# as Δ^3φ/ΔΩ^3 from Δ^3φ/ΔΩ^3 in (fs)^3\n",
    "TOD_foc = 0. * (1.e-15)**3"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "## Calculate value of requirement to neglect third order dispersion in pulse propagation, along central beam axis and at Rayleigh length\n",
    "z_R = sc.pi * focal_width**2 / wavel\n",
    "Ω_0 = 2.*sc.pi*sc.c/wavel\n",
    "SD_z = -AD_foc[4]*z_R\n",
    "β_3 = -SD_z/focal_width\n",
    "β_4 = β_3/Ω_0\n",
    "print(\"128 β_3 β_4/τ_0^3 = %e\"%(128*β_3*β_4/pulse_dur**3))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (2) Plot progression of the pulse's (dispersion) parameters while propagating through the focus"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "## Definitions for the plotting routines\n",
    "cmap = mpl.cm.gist_rainbow #mpl.cm.winter\n",
    "\n",
    "norm = mpl.colors.LogNorm(vmin=AD_mir[0] / dλdΩ * 1.e-3, vmax=AD_mir[-1] / dλdΩ * 1.e-3)\n",
    "\n",
    "cticks = AD_mir / dλdΩ * 1.e-3\n",
    "\n",
    "clabel = \"$AD_{in}$ [µrad/nm]\"\n",
    "\n",
    "colors = [cmap( norm(AD_mir_i / dλdΩ * 1.e-3) ) for AD_mir_i in AD_mir]\n",
    "\n",
    "z = Rayleigh_length * linspace(-5., 1., 121, endpoint=True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "## Create list of Fourier space pulses pulses\n",
    "ListOfFourierSpaceLasers = [\n",
    "    GaussPulseFourierSpace(\n",
    "        wavelength = wavel\n",
    "        , fourier_limited_pulse_duration = pulse_dur\n",
    "        , fourier_limited_width_in_focus = focal_width\n",
    "        , angular_dispersion_in_focus = AD_foc[i]\n",
    "        , spatial_dispersion_in_focus = SD_foc[i]\n",
    "        , group_delay_dispersion_in_focus = GDD_foc\n",
    "        , third_order_dispersion_in_focus = TOD_foc\n",
    "    )\n",
    "    for i in arange(len(AD_foc))\n",
    "]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "figure(figsize=(17,6))\n",
    "\n",
    "ax1 = subplot(1,2,1)\n",
    "for i, color in enumerate(colors):\n",
    "    Tanψ_tilt = PulseFrontTiltAnalytically(ListOfFourierSpaceLasers[i])\n",
    "    plot( z / Rayleigh_length, arctan(Tanψ_tilt(z)) * 180. / sc.pi, color = color )\n",
    "\n",
    "xlabel(r\"$z$ [$z_R$]\")\n",
    "xlim(z[0] / Rayleigh_length, z[-1] / Rayleigh_length)\n",
    "title(\"Pulse-front tilt [deg]\")\n",
    "grid()\n",
    "\n",
    "\n",
    "ax2 = subplot(1,2,2)\n",
    "for i, color in enumerate(colors):\n",
    "    Duration = PulseDurationAnalytically(ListOfFourierSpaceLasers[i])\n",
    "    plot( z / Rayleigh_length, Duration(z) / pulse_dur, color = color )\n",
    "\n",
    "xlabel(r\"$z$ [$z_R$]\")\n",
    "xlim(z[0] / Rayleigh_length, z[-1] / Rayleigh_length)\n",
    "title(\"Pulse duration [τ0]\")\n",
    "grid()\n",
    "\n",
    "\n",
    "cbar = colorbar(\n",
    "    mpl.cm.ScalarMappable(norm=norm, cmap=cmap)\n",
    "    , ax = [ax1,ax2]\n",
    "    , label = clabel\n",
    "    , ticks = cticks\n",
    ")\n",
    "\n",
    "cbar.ax.set_yticklabels([ \"%.1e\"%(cticks_i) for cticks_i in cticks ])\n",
    "\n",
    "suptitle(r\"$f/%.1f$ OAP, $z_R=%.0f$µm, $SD_{in}=%.0f$µm/nm\"%(f_num, Rayleigh_length * 1.e6, SD_mir[0] / dλdΩ * 1.e6 / 1.e9))\n",
    "\n",
    "show()\n",
    "\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## (3) Plot inset showing that PFT != 0 in focus"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "figure(figsize=(4,3))\n",
    "\n",
    "z_zoom = Rayleigh_length * linspace(-0.005, 0.005, 201, endpoint=True)\n",
    "\n",
    "ax1 = subplot(1,1,1)\n",
    "for i, color in enumerate(colors):\n",
    "    Tanψ_tilt = PulseFrontTiltAnalytically(ListOfFourierSpaceLasers[i])\n",
    "    plot( z_zoom * 1.e6, arctan(Tanψ_tilt(z_zoom)) * 180. / sc.pi, color = color )\n",
    "    \n",
    "xlabel(r\"$z$ [µm]\")\n",
    "xlim(-0.05, 0.05)\n",
    "xticks(array([-0.05, 0., 0.05]))\n",
    "ylim(-.05, .05)\n",
    "yticks(array([-0.05, 0., 0.05]))\n",
    "grid()\n",
    "\n",
    "show()\n",
    "#savefig(\"pics/DRACO_PFT-PD-scaling_short-range_w-o-SDin_zoom.png\", dpi=100, transparent=False, bbox_inches='tight')\n",
    "\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Form of pulse width function\n",
    "\n",
    "$$\n",
    "E \\propto e^{-\\frac{x^2}{W^2}}\\,,\\text{ where }\n",
    "  W^2 = w^2 \\frac{\\left(1 + 2\\frac{x}{w}\\frac{2}{\\Omega_0\\tau_0} \\frac{2\\theta^\\prime z / \\tau_0}{w} + \\left( \\frac{2\\theta^\\prime z / \\tau_0}{w} \\right)^2 \\right)}\n",
    "    {\\left(1 + 2\\frac{x}{w}\\frac{2}{\\Omega_0\\tau_0} \\frac{2\\theta^\\prime z / \\tau_0}{w} \\right)}\\,\\text{ for } SD_0 = 0\\,.\n",
    "$$\n",
    "\n",
    "If $W$ is to keep its meaning of the width of the transverse profile, which it has for a Gaussian Pulse, i.e. in case $(2\\theta^\\prime z)/(\\tau_0 w) \\ll 1$, then $x^2/W^2 \\rvert_{x=W}=1$.\n",
    "That is,\n",
    "$$\n",
    "\\frac{W^2}{w^2} \\frac{\\left(1 + 2\\frac{W}{w}\\frac{2}{\\Omega_0\\tau_0} \\frac{2\\theta^\\prime z / \\tau_0}{w} \\right)}\n",
    "  {\\left(1 + 2\\frac{W}{w}\\frac{2}{\\Omega_0\\tau_0} \\frac{2\\theta^\\prime z / \\tau_0}{w} + \\left( \\frac{2\\theta^\\prime z / \\tau_0}{w} \\right)^2 \\right)} \\stackrel{!}{=} 1\n",
    "$$\n",
    "which yields a cubic equation in $W/w$.\n",
    "$$\n",
    "\\Rightarrow 0 \\stackrel{!}{=} 2\\frac{W^3}{w^3}\\frac{2}{\\Omega_0\\tau_0} \\frac{2\\theta^\\prime z / \\tau_0}{w}\n",
    "  + \\frac{W^2}{w^2}\n",
    "  - 2\\frac{W}{w}\\frac{2}{\\Omega_0\\tau_0} \\frac{2\\theta^\\prime z / \\tau_0}{w}\n",
    "  - \\left(1 + \\left( \\frac{2\\theta^\\prime z / \\tau_0}{w} \\right)^2 \\right)\n",
    "$$\n",
    "$$\n",
    "\\Rightarrow 0 = Ax^3 + Bx^2 + Cx + D\\,,\\text{ where } C=-A\\,, B=1\\,, D<0\n",
    "$$\n",
    "$$\n",
    "\\Rightarrow 0 = (x^3 - x) A + x^2 + D\n",
    "$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "x = linspace(1, 3.5, 100) # = W/w\n",
    "Ωτ = 10.\n",
    "\n",
    "rel_walk_off = 10**linspace(-1,1,3)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "print(rel_walk_off)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "figure()\n",
    "for rel_walk_off_i in rel_walk_off:\n",
    "    plot(x, (x**3-x)*(rel_walk_off_i*4/Ωτ) + x**2 - 1 - rel_walk_off_i**2, label=\"%s\"%(rel_walk_off_i))\n",
    "xlabel(\"W/w\")\n",
    "ylabel(r\"Defining equation for W/w\")\n",
    "grid()\n",
    "legend()\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Test implementations with standard Gauss"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true
   },
   "source": [
    "## Full spectral field"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true,
    "tags": []
   },
   "outputs": [],
   "source": [
    "FourierGaus = GaussPulseFourierSpace(\n",
    "    wavelength = 800.e-9 # wavelength of the central laser frequency, unit: m\n",
    "    , fourier_limited_pulse_duration = 30.e-15 / 1.17741 # tau_FWHM,I / sqrt(2. * log(2.))\n",
    "    , fourier_limited_width_in_focus = 2.5e-6 / 1.17741 # w_FWHM,I / sqrt(2. * log(2.))\n",
    "    , angular_dispersion_in_focus = 0. # dθ/dΩ, e.g. tan(α_tilt)/Ω0, unit: rad * s\n",
    "    , spatial_dispersion_in_focus = 0. # dx/dΩ, unit: m*s\n",
    "    , group_delay_dispersion_in_focus = -1000. * (1.e-15)**2 # (d^2φ/dΩ^2) unit: s^2\n",
    "    , third_order_dispersion_in_focus = 0. * (1.e-15)**3 # (d^3φ/dΩ^3) unit: s^3\n",
    ")\n",
    "RealGaus = RealSpaceLaserPulse(FourierGaus, N_λ = 8., N_T = 20.) #N_T = 7.)\n",
    "IntensGaus = RealSpaceIntensityEnvelope(FourierGaus, N_λ = 8., N_T = 20.) #N_T = 7.)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Check spectrum of gaus in Fourier space"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true,
    "tags": []
   },
   "outputs": [],
   "source": [
    "figure()\n",
    "\n",
    "plot((RealGaus.ν/FourierGaus.ν0 - 1.), FourierGaus(RealGaus.ν, 0., 0.))\n",
    "xlabel(\"$(\\Omega-\\Omega_0)/\\Omega_0$\")\n",
    "ylabel(\"$\\hat E$\")\n",
    "\n",
    "grid()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "figure()\n",
    "\n",
    "plot((IntensGaus.ν/FourierGaus.ν0 - 1.), FourierGaus(IntensGaus.ν, 0., 0.))\n",
    "\n",
    "xlabel(\"$(\\Omega-\\Omega_0)/\\Omega_0$\")\n",
    "ylabel(\"$\\hat E$\")\n",
    "\n",
    "grid()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Check field calculation by comparing to expected field and envelope at x=z=0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true,
    "tags": []
   },
   "outputs": [],
   "source": [
    "figure(figsize=(16,12))\n",
    "plot(RealGaus.t/FourierGaus.τ0, RealGaus(0., 0.,), label=\"DFT\")\n",
    "plot(RealGaus.t/FourierGaus.τ0, exp(- (RealGaus.t/FourierGaus.τ0)**2 ), label=\"field envelope\")\n",
    "plot(RealGaus.t/FourierGaus.τ0\n",
    "     , exp(- (RealGaus.t/FourierGaus.τ0)**2 )*cos(2.*sc.pi*FourierGaus.ν0*RealGaus.t)\n",
    "     , '.', ms=12,label=\"Gauss field\")\n",
    "xlabel(\"t [τ_Laser]\")\n",
    "ylabel(\"E\")\n",
    "\n",
    "grid()\n",
    "\n",
    "legend()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Check intensity calculation by comparing to expected intensity at x=z=0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true,
    "tags": []
   },
   "outputs": [],
   "source": [
    "figure(figsize=(16,12))\n",
    "plot(IntensGaus.t/FourierGaus.τ0, IntensGaus(0., 0.,), label=\"DFT\")\n",
    "#plot(IntensGaus.t/FourierGaus.τ0, exp(- 2.*(IntensGaus.t/FourierGaus.τ0)**2 )\n",
    "#     , \".\", label=\"intensity envelope\")\n",
    "xlabel(\"t [τ0]\")\n",
    "ylabel(\"I\")\n",
    "\n",
    "ylim(0., 0.5)\n",
    "\n",
    "grid()\n",
    "\n",
    "legend()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Calculate field in (t,x) domain"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "x = linspace(-2.*FourierGaus.w0, 2.*FourierGaus.w0, 65)\n",
    "z = FourierGaus.zR * linspace(0., 1., 5, endpoint=True)\n",
    "\n",
    "print(\"z ϵ [%.2f, %.2f]z_R\"%(z[0]/FourierGaus.zR, z[-1]/FourierGaus.zR))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "for z_i in z:\n",
    "    \n",
    "    t_norm = FourierGaus.τ0\n",
    "    x_norm = FourierGaus.w0\n",
    "    \n",
    "    field2D_t = array([ RealGaus(x_i, z_i) for x_i in x ])\n",
    "\n",
    "    intens = abs(field2D_t)**2\n",
    "\n",
    "    x_center = len(x)//2\n",
    "    max_center = argmax(intens[x_center])\n",
    "    x_out = int(x_center + 1.*FourierGaus.w0 / abs(x[1]-x[0]) )\n",
    "    max_out = argmax(intens[x_out])\n",
    "    \n",
    "    figure(figsize=(16,6))\n",
    "    \n",
    "    print(\"Verkippung = arctan( (%.2f - %.2f  ) / ( %.2f - %.2f  ) ) = %.2f deg\"%(\n",
    "        RealGaus.t[max_center] / t_norm\n",
    "        , RealGaus.t[max_out] / t_norm\n",
    "        , x[x_center] / (sc.c*t_norm)\n",
    "        , x[x_out] / (sc.c*t_norm)\n",
    "        , arctan((sc.c*RealGaus.t[max_center] - sc.c*RealGaus.t[max_out]) / (x[x_center] - x[x_out]))*180./pi)\n",
    "    )\n",
    "    \n",
    "    subplot(1,2,1)\n",
    "    \n",
    "    imshow(intens,cmap = mpl.cm.Reds\n",
    "           , origin='lower'\n",
    "           , interpolation='none', aspect='auto'\n",
    "           , extent=(RealGaus.t[0] / t_norm, RealGaus.t[-1] / t_norm\n",
    "                     , x[0] / x_norm, x[-1] / x_norm )\n",
    "           #, norm=matplotlib.colors.LogNorm(vmin=1.e-0)\n",
    "          )\n",
    "\n",
    "    colorbar()\n",
    "    \n",
    "    xlabel(r\"$t$ [$\\tau_0$]\")\n",
    "    ylabel(r\"$x$ [$w_0$]\")\n",
    "        \n",
    "    title(\"z = %.2f z_R\"%(z_i /  FourierGaus.zR))\n",
    "    \n",
    "    grid()\n",
    "        \n",
    "    show()\n",
    "    \n",
    "    close()\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true
   },
   "source": [
    "## Approximated spectral field\n",
    "\n",
    "Using the field that is actually Fourier transformed for the analytic calculation.\n",
    "That is, the field with highest order of 2 in (Ω-Ω0)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "ApproxFourierGaus = ApproximateGaussPulseFourierSpace(\n",
    "    wavelength = 800.e-9 # wavelength of the central laser frequency, unit: m\n",
    "    , fourier_limited_pulse_duration = 5.e-15 / 1.17741 # tau_FWHM,I / sqrt(2. * log(2.))\n",
    "    , fourier_limited_width_in_focus = 3.e-6 / 1.17741 # w_FWHM,I / sqrt(2. * log(2.))\n",
    "    , angular_dispersion_in_focus = 0. # dθ/dΩ, e.g. tan(α_tilt)/Ω0, unit: rad * s\n",
    "    , spatial_dispersion_in_focus = 0. # dx/dΩ, unit: m*s\n",
    "    , group_delay_dispersion_in_focus = 0. # (d^2φ/dΩ^2) unit: s^2\n",
    ")\n",
    "ApproxRealGaus = RealSpaceLaserPulse(ApproxFourierGaus, N_λ = 8., N_T = 7.)\n",
    "ApproxIntensGaus = RealSpaceIntensityEnvelope(ApproxFourierGaus, N_λ = 8., N_T = 7.)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Check spectrum of gaus in Fourier space"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "figure()\n",
    "\n",
    "plot((ApproxRealGaus.ν/ApproxFourierGaus.ν0 - 1.), ApproxFourierGaus(ApproxRealGaus.ν, 0., 0.))\n",
    "xlabel(\"$(\\Omega-\\Omega_0)/\\Omega_0$\")\n",
    "ylabel(\"$\\hat E$\")\n",
    "\n",
    "grid()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "figure()\n",
    "\n",
    "plot((ApproxIntensGaus.ν/ApproxFourierGaus.ν0 - 1.), ApproxFourierGaus(ApproxIntensGaus.ν, 0., 0.))\n",
    "\n",
    "xticks(arange(-5, 3+1))\n",
    "\n",
    "xlabel(\"$(\\Omega-\\Omega_0)/\\Omega_0$\")\n",
    "ylabel(\"$\\hat E$\")\n",
    "\n",
    "grid()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Check field calculation by comparing to expected field and envelope at x=z=0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "figure(figsize=(16,12))\n",
    "plot(ApproxRealGaus.t/ApproxFourierGaus.τ0, ApproxRealGaus(0., 0.,), label=\"DFT\")\n",
    "plot(ApproxRealGaus.t/ApproxFourierGaus.τ0, exp(- (ApproxRealGaus.t/ApproxFourierGaus.τ0)**2 ), label=\"field envelope\")\n",
    "plot(ApproxRealGaus.t/ApproxFourierGaus.τ0\n",
    "     , exp(- (ApproxRealGaus.t/ApproxFourierGaus.τ0)**2 )*cos(2.*sc.pi*ApproxFourierGaus.ν0*ApproxRealGaus.t)\n",
    "     , '.', ms=12,label=\"Gauss field\")\n",
    "xlabel(\"t [τ_Laser]\")\n",
    "ylabel(\"E\")\n",
    "\n",
    "grid()\n",
    "\n",
    "legend()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Check intensity calculation by comparing to expected intensity at x=z=0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "figure()\n",
    "plot(ApproxIntensGaus.t/ApproxFourierGaus.τ0, ApproxIntensGaus(0., 0.,), label=\"DFT\")\n",
    "plot(ApproxIntensGaus.t/ApproxFourierGaus.τ0, exp(- 2.*(ApproxIntensGaus.t/ApproxFourierGaus.τ0)**2 )\n",
    "     , \".\", label=\"intensity envelope\")\n",
    "xlabel(\"t [τ_Laser]\")\n",
    "ylabel(\"I\")\n",
    "\n",
    "grid()\n",
    "\n",
    "legend()\n",
    "\n",
    "show()\n",
    "close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "### Calculate field in (t,x) domain"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "x = linspace(-2.*ApproxFourierGaus.w0, 2.*ApproxFourierGaus.w0, 64)\n",
    "z = ApproxFourierGaus.zR * linspace(0., 1., 5, endpoint=True)\n",
    "\n",
    "print(\"z ϵ [%.2f, %.2f]z_R\"%(z[0]/ApproxFourierGaus.zR, z[-1]/ApproxFourierGaus.zR))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "for z_i in z:\n",
    "    \n",
    "    t_norm = ApproxFourierGaus.τ0\n",
    "    x_norm = ApproxFourierGaus.w0\n",
    "    \n",
    "    intens2D_t = array([ ApproxIntensGaus(x_i, z_i) for x_i in x ])\n",
    "\n",
    "    x_center = len(x)//2\n",
    "    max_center = argmax(intens2D_t[x_center])\n",
    "    x_out = int(x_center + 1.*ApproxFourierGaus.w0 / abs(x[1]-x[0]) )\n",
    "    max_out = argmax(intens2D_t[x_out])\n",
    "    \n",
    "    figure(figsize=(16,6))\n",
    "    \n",
    "    print(\"Verkippung = arctan( (%.2f - %.2f  ) / ( %.2f - %.2f  ) ) = %.2f deg\"%(\n",
    "        ApproxRealGaus.t[max_center] / t_norm\n",
    "        , ApproxRealGaus.t[max_out] / t_norm\n",
    "        , x[x_center] / (sc.c*t_norm)\n",
    "        , x[x_out] / (sc.c*t_norm)\n",
    "        , arctan((sc.c*ApproxRealGaus.t[max_center] - sc.c*ApproxRealGaus.t[max_out]) / (x[x_center] - x[x_out]))*180./pi)\n",
    "    )\n",
    "    \n",
    "    subplot(1,2,1)\n",
    "    \n",
    "    imshow(intens2D_t,cmap = mpl.cm.Reds\n",
    "           , origin='lower'\n",
    "           , interpolation='none', aspect='auto'\n",
    "           , extent=(ApproxRealGaus.t[0] / t_norm, ApproxRealGaus.t[-1] / t_norm\n",
    "                     , x[0] / x_norm, x[-1] / x_norm)\n",
    "           #, norm=matplotlib.colors.LogNorm(vmin=1.e-0)\n",
    "          )\n",
    "\n",
    "    colorbar()\n",
    "    \n",
    "    xlabel(r\"$t$ [$\\tau_0$]\")\n",
    "    ylabel(r\"$x$ [$w_0$]\")\n",
    "        \n",
    "    title(\"z = %.2f z_R\"%(z_i /  ApproxFourierGaus.zR))\n",
    "    \n",
    "    grid()\n",
    "        \n",
    "    show()\n",
    "    \n",
    "    close()\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true
   },
   "source": [
    "# Gaus with tilt"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "wavel = 800.e-9\n",
    "FourierGausWithAD = GaussPulseFourierSpace(\n",
    "    wavelength = wavel\n",
    "    , fourier_limited_pulse_duration = 5.e-15 / 1.17741 # tau_FWHM,I / sqrt(2. * log(2.))\n",
    "    , fourier_limited_width_in_focus = 3.e-6 / 1.17741 # w_FWHM,I / sqrt(2. * log(2.))\n",
    "    , angular_dispersion_in_focus = -tan(30.*sc.pi/180.)/(2.*sc.pi*sc.c/wavel)\n",
    "    , spatial_dispersion_in_focus = 0.\n",
    "    , group_delay_dispersion_in_focus = 0.\n",
    ")\n",
    "RealGausWithAD = RealSpaceLaserPulse(FourierGausWithAD, N_λ = 4., N_T = 7.)\n",
    "IntensGausWithAD = RealSpaceIntensityEnvelope(FourierGausWithAD, N_λ = 4., N_T = 7.)\n",
    "Tanψ_tilt = PulseFrontTiltAnalytically(FourierGausWithAD)\n",
    "Front = PulseFront(FourierGausWithAD)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "## Calculate field in (t,x) domain"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "x = linspace(-2.*FourierGausWithAD.w0, 2.*FourierGausWithAD.w0, 65)\n",
    "z = FourierGausWithAD.zR * linspace(-5., 0., 11, endpoint=True)\n",
    "\n",
    "print(\"z ϵ [%.2f, %.2f]z_R\"%(z[0]/FourierGausWithAD.zR, z[-1]/FourierGausWithAD.zR))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "for z_i in z:\n",
    "    \n",
    "    t_norm = FourierGausWithAD.τ0\n",
    "    x_norm = FourierGausWithAD.w0\n",
    "    \n",
    "    intens2D_t = array([ IntensGausWithAD(x_i, z_i) for x_i in x ])\n",
    "\n",
    "    ## Calculate PFT from field distribution calculated with DFT\n",
    "    x_center = len(x-1)//2\n",
    "    max_center = argmax(intens2D_t[x_center])\n",
    "    x_out = int(x_center + 1.*FourierGausWithAD.w0 / abs(x[1]-x[0]) )\n",
    "    max_out = argmax(intens2D_t[x_out])\n",
    "    \n",
    "    tanψ_tilt_num = ( (sc.c*IntensGausWithAD.t[max_center] - sc.c*IntensGausWithAD.t[max_out])\n",
    "                      / (x[x_center] - x[x_out]) )\n",
    "    print(\"PFT (num) = arctan( (%.2f - %.2f  ) / ( %.2f - %.2f  ) ) = %.2f deg\"%(\n",
    "        IntensGausWithAD.t[max_center] / t_norm\n",
    "        , IntensGausWithAD.t[max_out] / t_norm\n",
    "        , x[x_center] / x_norm\n",
    "        , x[x_out] / x_norm\n",
    "        , arctan(tanψ_tilt_num) * 180. / pi\n",
    "    )\n",
    "    )\n",
    "    \n",
    "    print(\"PFT (ana) = %.2f deg\"%( arctan(Tanψ_tilt(z_i)) * 180. / sc.pi ))\n",
    "    \n",
    "    figure(figsize=(8,6))\n",
    "    \n",
    "    subplot(1,1,1)\n",
    "    \n",
    "    imshow(intens2D_t,cmap = mpl.cm.Reds\n",
    "           , origin='lower'\n",
    "           , interpolation='none', aspect='auto'\n",
    "           , extent=(IntensGausWithAD.t[0] / t_norm\n",
    "                     , IntensGausWithAD.t[-1] / t_norm\n",
    "                     , x[0] / x_norm\n",
    "                     , x[-1] / x_norm\n",
    "                    )\n",
    "           #, norm=matplotlib.colors.LogNorm(vmin=1.e-0)\n",
    "          )\n",
    "\n",
    "    colorbar()\n",
    "    \n",
    "    ## Plot theoretical pulse-front contour.\n",
    "    ## Since I plot t on the x axis, and x ond the y axis, the slope of the pulse-front is 90°-ψ_tilt.\n",
    "    ## Here I use tan( 90°-ψ_tilt ) = 1 / tan( ψ_tilt )\n",
    "    plot(Tanψ_tilt(z_i) * x / (sc.c * t_norm), x / x_norm, label=\"analytic PFT\" )\n",
    "    plot(tanψ_tilt_num * x / (sc.c * t_norm), x / x_norm, label=\"measured PFT\" )\n",
    "    plot(array([Front(x_i, z_i) for x_i in x ]) / t_norm, x / x_norm, label=\"analytic Front\" )\n",
    "    \n",
    "    xlabel(r\"$t$ [$\\tau_0$]\")\n",
    "    ylabel(r\"$x$ [$w_0$]\")\n",
    "    \n",
    "    title(\"Intensity @ z = %.2f z_R\"%(z_i /  FourierGausWithAD.zR))\n",
    "    \n",
    "    grid()\n",
    "    \n",
    "    legend()\n",
    "        \n",
    "    show()\n",
    "    \n",
    "    close()\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "heading_collapsed": true,
    "hidden": true
   },
   "source": [
    "## Aproximate Gauss pulse with AD"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "wavel = 800.e-9\n",
    "ApproxFourierGausWithAD = ApproximateGaussPulseFourierSpace(\n",
    "    wavelength = wavel\n",
    "    , fourier_limited_pulse_duration = 5.e-15 / 1.17741 # tau_FWHM,I / sqrt(2. * log(2.))\n",
    "    , fourier_limited_width_in_focus = 3.e-6 / 1.17741 # w_FWHM,I / sqrt(2. * log(2.))\n",
    "    , angular_dispersion_in_focus = tan(30.*sc.pi/180.)/(2.*sc.pi*sc.c/wavel)\n",
    "    , spatial_dispersion_in_focus = 0.\n",
    "    , group_delay_dispersion_in_focus = 0.\n",
    ")\n",
    "ApproxRealGausWithAD = RealSpaceLaserPulse(ApproxFourierGausWithAD, N_λ = 12., N_T = 7.)\n",
    "ApproxIntensGausWithAD = RealSpaceIntensityEnvelope(ApproxFourierGausWithAD, N_λ = 12., N_T = 7.)\n",
    "ApproxTanψ_tilt = PulseFrontTiltAnalytically(ApproxFourierGausWithAD)\n",
    "ApproxFront = PulseFront(ApproxFourierGausWithAD)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "x = sc.c * ApproxRealGausWithAD.t\n",
    "z = ApproxFourierGausWithAD.zR * linspace(0., 1., 5, endpoint=True)\n",
    "\n",
    "print(\"z ϵ [%.2f, %.2f]z_R\"%(z[0]/ApproxFourierGausWithAD.zR, z[-1]/ApproxFourierGausWithAD.zR))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "hidden": true
   },
   "outputs": [],
   "source": [
    "for z_i in z:\n",
    "    \n",
    "    t_norm = ApproxFourierGausWithAD.τ0\n",
    "    x_norm = ApproxFourierGausWithAD.w0\n",
    "    \n",
    "    intens2D_t = array([ ApproxIntensGausWithAD(x_i, z_i) for x_i in x ]) #array([ IntensGausWithAD(x_i, z_i) for x_i in x ])\n",
    "\n",
    "    ## Calculate PFT from field distribution calculated with DFT\n",
    "    x_center = len(x-1)//2\n",
    "    max_center = argmax(intens2D_t[x_center])\n",
    "    x_out = int(x_center + 1.*ApproxFourierGausWithAD.w0 / abs(x[1]-x[0]) )\n",
    "    max_out = argmax(intens2D_t[x_out])\n",
    "        \n",
    "    tanψ_tilt_num = ( (sc.c*ApproxIntensGausWithAD.t[max_center] - sc.c*ApproxIntensGausWithAD.t[max_out])\n",
    "                      / (x[x_center] - x[x_out]) )\n",
    "    print(\"Verkippung (num) = arctan( (%.2f - %.2f  ) / ( %.2f - %.2f  ) ) = %.2f deg\"%(\n",
    "        ApproxIntensGausWithAD.t[max_center] / t_norm\n",
    "        , ApproxIntensGausWithAD.t[max_out] / t_norm\n",
    "        , x[x_center] / x_norm\n",
    "        , x[x_out] / x_norm\n",
    "        , arctan(tanψ_tilt_num) * 180. / pi\n",
    "    )\n",
    "    )\n",
    "    \n",
    "    print(\"Verkippung (ana) = %.2f deg\"%( arctan(ApproxTanψ_tilt(z_i)) * 180. / sc.pi ))\n",
    "    \n",
    "    figure(figsize=(8,6))\n",
    "\n",
    "    subplot(1,1,1)\n",
    "    \n",
    "    imshow(intens2D_t,cmap = mpl.cm.Reds\n",
    "           , origin='lower'\n",
    "           , interpolation='none', aspect='auto'\n",
    "           , extent=(ApproxIntensGausWithAD.t[0] / t_norm\n",
    "                     , ApproxIntensGausWithAD.t[-1] / t_norm\n",
    "                     , x[0] / x_norm\n",
    "                     , x[-1] / x_norm\n",
    "                    )\n",
    "           #, norm=matplotlib.colors.LogNorm(vmin=1.e-0)\n",
    "          )\n",
    "\n",
    "    colorbar()\n",
    "    \n",
    "    ## Plot theoretical pulse-front contour.\n",
    "    ## Since I plot t on the x axis, and x ond the y axis, the slope of the pulse-front is 90°-ψ_tilt.\n",
    "    ## Here I use tan( 90°-ψ_tilt ) = 1 / tan( ψ_tilt )\n",
    "    plot(ApproxIntensGausWithAD.t / t_norm, ( 1. / ApproxTanψ_tilt(z_i) ) * sc.c * ApproxIntensGausWithAD.t / x_norm, label=\"analytic PFT\" )\n",
    "    plot(ApproxIntensGausWithAD.t / t_norm, ( 1. / tanψ_tilt_num ) * sc.c * ApproxIntensGausWithAD.t / x_norm, label=\"measured PFT\" )\n",
    "    plot(array([ApproxFront(x_i, z_i) for x_i in x ]) / t_norm, x / x_norm, label=\"analytic Front\" )\n",
    "    \n",
    "    ylim(x[0] / x_norm, x[-1] / x_norm)\n",
    "    \n",
    "    xlabel(r\"$t$ [$\\tau_0$]\")\n",
    "    ylabel(r\"$x$ [$w_0$]\")\n",
    "    \n",
    "    title(\"Intensity @ z = %.2f z_R\"%(z_i /  ApproxFourierGausWithAD.zR))\n",
    "    \n",
    "    grid()\n",
    "    \n",
    "    legend()\n",
    "        \n",
    "    show()\n",
    "    \n",
    "    close()\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.12"
  },
  "toc": {
   "base_numbering": 1,
   "nav_menu": {},
   "number_sections": true,
   "sideBar": true,
   "skip_h1_title": false,
   "title_cell": "Table of Contents",
   "title_sidebar": "Contents",
   "toc_cell": false,
   "toc_position": {},
   "toc_section_display": true,
   "toc_window_display": false
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
